# =======================================
# #         Emulator                    #
# =======================================


class EmulatorEngine(object):

    def __init__(self, instructions):
        """ TODO """
        raise NotImplementedError

    def emulate(self, state, depth=0):
        """ TODO """
        raise NotImplementedError

    def emulate_one_instruction(self, instr, state, depth):
        """ TODO """
        raise NotImplementedError


from octopus.arch.wasm.disassembler import WasmDisassembler


# Eos smart contract == wasm module
class EosDisassembler(WasmDisassembler):
    def __init__(self, bytecode=None):
        WasmDisassembler.__init__(self,
                                  bytecode=bytecode)



from binascii import unhexlify


def bytecode_to_bytes(bytecode):
    if str(bytecode).startswith("0x"):
        bytecode = bytecode[2:]

    try:
        # python > 2.7
        bytecode = bytes.fromhex(bytecode)
    except AttributeError:
        # python <= 2.7
        try:
            bytecode = bytecode.decode("hex")
        except TypeError:
            # last chance
            try:
                bytecode = unhexlify(bytecode)
            except:
                pass
    # already bytes or bytearray
    except TypeError:
        pass
    return bytecode


def search_in_list_of_dict(string_to_search, target_list, key_dict):
    return list(filter(lambda elem: str(string_to_search) in str(elem[key_dict]), target_list))



#from .constants import TT256, TT255

import re
from z3 import *

TT256 = 2 ** 256
TT256M1 = 2 ** 256 - 1
TT255 = 2 ** 255


class helper(object):

    def is_symbolic(value):
        return not isinstance(value, int)

    def is_real(value):
        return isinstance(value, int)

    def isAllReal(*args):
        for i in args:
            if is_symbolic(i):
                return False
        return True

    def safe_decode(hex_encoded_string):

        if (hex_encoded_string.startswith("0x")):
            return bytes.fromhex(hex_encoded_string[2:])
        else:
            return bytes.fromhex(hex_encoded_string)

    def to_signed(i):
        return i if i < TT255 else i - TT256

    def get_trace_line(instr, state):

        stack = str(state.stack[::-1])

        # stack = re.sub("(\d+)",    lambda m: hex(int(m.group(1))), stack)
        stack = re.sub("\n", "", stack)

        return str(instr['address']) + " " + instr['opcode'] + "\tSTACK: " + stack

    def convert_to_bitvec(item):
        # converting boolean expression to bitvector
        if type(item) == BoolRef:
            return If(item, BitVecVal(1, 256), BitVecVal(0, 256))
        elif type(item) == bool:
            if item:
                return BitVecVal(1, 256)
            else:
                return BitVecVal(0, 256)
        elif type(item) == int:
            return BitVecVal(item, 256)
        else:
            return simplify(item)

    def convert_to_concrete_int(item):
        if (type(item) == int):
            return item

        if (type(item) == BitVecNumRef):
            return item.as_long()

        return simplify(item).as_long()

    def get_concrete_int(item):

        if (type(item) == int):
            return item

        if (type(item) == BitVecNumRef):
            return item.as_long()

        return simplify(item).as_long()

    def concrete_int_from_bytes(_bytes, start_index):

        # logging.debug("-- concrete_int_from_bytes: " + str(_bytes[start_index:start_index+32]))
        b = _bytes[start_index:start_index+32]

        val = int.from_bytes(b, byteorder='big')

        return val

    def concrete_int_to_bytes(val):

        # logging.debug("concrete_int_to_bytes " + str(val))

        if (type(val) == int):
            return val.to_bytes(32, byteorder='big')

        return (simplify(val).as_long()).to_bytes(32, byteorder='big')



from octopus.engine.explorer import Explorer
from octopus.platforms.BTC.bech32 import encode as bech32_encode

import binascii

RPC_USER = 'test'
RPC_PASSWORD = 'test'
RPC_HOST = 'localhost'

BITCOIND_DEFAULT_RPC_PORT = 8332

class BitcoinExplorerRPC(Explorer):
    '''
    Bitcoin JSON-RPC client class
    '''

    def __init__(self, host='localhost', port=BITCOIND_DEFAULT_RPC_PORT, tls=False, max_retries=3):
        Explorer.__init__(self, host=host, port=port, tls=tls, max_retries=max_retries)

    def call(self, method, params=None, jsonrpc='2.0', _id=None):
        return super().call(method, params, jsonrpc, _id)

    #######################
    # HIGHT-LEVEL METHODS #
    #######################

    def get_transaction(self, transaction_id, verbosity=1):
        """ Return transaction informations

        .. seealso::
            getrawtransaction()
        """
        return self.getrawtransaction(transaction_id, verbosity)

    def get_block_by_number(self, block_number):
        """ Return block information using given block number

        .. seealso::
            getblockhash() + getblock()
        """
        return self.getblock(self.getblockhash(block_number))

    def get_block_by_hash(self, block_hash):
        """ Return block information using given block hash

        .. seealso::
            getblock()
        """
        return self.getblock(block_hash)

    def decode_tx(self, tx):

        result = {"txid": tx['txid'],
                  "sender": [],
                  "receiver": []
                  }

        for vin in tx['vin']:
            # handle first transaction that's the generated bitcoin from the network

            try:
                coinbase = vin['coinbase']
                result['sender'].append(('mining_reward', 0))
            except KeyError:
                # KeyError appears when vin['coinbase'] doesn't exit
                # following is common transaction between 2 bitcoin wallet

                # search for sender address
                sender_txid = vin['txid']
                sender_num = vin['vout']
                sender_tx = self.getrawtransaction(sender_txid, 1)
                for sender_vout in sender_tx['vout']:
                    if sender_vout['n'] == sender_num:
                        try:
                            result['sender'].append((sender_vout['scriptPubKey']['addresses'][0], sender_vout['value']))
                        except KeyError:
                            # Native segwit outputs (P2WSH and P2WPKH) do not currently have an address
                            # so there are no addresses for these outputs.
                            if sender_vout['scriptPubKey']['type'] in ['witness_v0_keyhash', 'witness_v0_scripthash']:
                                result['sender'].append(('Native_SegWit', sender_vout['value']))
                            elif sender_vout['scriptPubKey']['type'] in ['nonstandard', 'pubkey']:
                                result['sender'].append((sender_vout['scriptPubKey']['type'], sender_vout['value']))
                            else:
                                raise Exception('decoding sender impossible : %s', sender_tx)

        # association between sender, receiver and amount
        for vout in tx['vout']:
            value = vout['value']
            if value != 0.0:
                try:
                    receiver = vout['scriptPubKey']['addresses'][0]
                    btcscript = vout['scriptPubKey']['hex']
                    result['receiver'].append((receiver, value, btcscript))
                except KeyError:
                    # test if Native SegWit outputs
                    if vout['scriptPubKey']['type'] in ['witness_v0_keyhash', 'witness_v0_scripthash']:
                        # Native segwit outputs (P2WSH and P2WPKH) do not currently have an address type
                        # so there are no addresses for these outputs.
                        receiver = 'Native_SegWit'
                        btcscript = vout['scriptPubKey']['hex']
                        result['receiver'].append((receiver, value, btcscript))
                    elif vout['scriptPubKey']['type'] in ['nonstandard', 'pubkey']:
                        receiver = vout['scriptPubKey']['type']
                        btcscript = vout['scriptPubKey']['hex']
                        result['receiver'].append((receiver, value, btcscript))
                    # unknown error
                    else:
                        raise Exception('decoding receiver impossible : %s', tx)

        return result

    ####################
    # JSON-RPC METHODS #
    ####################

    # ressources :
    # * https://en.bitcoin.it/wiki/Original_Bitcoin_client/API_calls_list
    # * https://en.bitcoin.it/wiki/Elis-API
    # * http://chainquery.com/bitcoin-api

    def abandontransaction(self, txid):
        '''
        http://chainquery.com/bitcoin-api/abandontransaction

        NOT TESTED

        Mark in-wallet transaction <txid> as abandoned
        This will mark this transaction and all its in-wallet descendants as abandoned which will allow
        for their inputs to be respent.  It can be used to replace "stuck" or evicted transactions.
        It only works on transactions which are not included in a block and are not currently in the mempool.
        It has no effect on transactions which are already conflicted or abandoned.

        Arguments:
        1. "txid"    (string, required) The transaction id

        Result: Null
        '''
        return self.call('abandontransaction', [txid])

    def addmultisigaddress(self, nrequired, keys):
        '''
        http://chainquery.com/bitcoin-api/addmultisigaddress

        NOT TESTED

        Add a nrequired-to-sign multisignature address to the wallet.
        Each key is a Bitcoin address or hex-encoded public key.
        If 'account' is specified (DEPRECATED), assign address to that account.

        Arguments:
        1. nrequired        (numeric, required) The number of required signatures out of the n keys or addresses.
        2. "keys"         (string, required) A json array of bitcoin addresses or hex-encoded public keys
             [
               "address"  (string) bitcoin address or hex-encoded public key
               ...,
             ]
        3. "account"      (string, optional) DEPRECATED. An account to assign the addresses to.

        Result:
        "address"         (string) A bitcoin address associated with the keys.
        '''
        return self.call('addmultisigaddress'[nrequired, keys])

    def addnode(self, node, command):
        '''
        http://chainquery.com/bitcoin-api/addnode

        NOT TESTED

        Attempts to add or remove a node from the addnode list.
        Or try a connection to a node once.

        Arguments:
        1. "node"     (string, required) The node (see getpeerinfo for nodes)
        2. "command"  (string, required) 'add' to add a node to the list, 'remove' to remove a node from the list, 'onetry' to try a connection to the node once

        '''
        return self.call('addnode', [node, command])

    def createmultisig(self, nrequired, keys):
        '''
        http://chainquery.com/bitcoin-api/createmultisig

        NOT TESTED

        Creates a multi-signature address with n signature of m keys required.
        It returns a json object with the address and redeemScript.

        Arguments:
        1. nrequired      (numeric, required) The number of required signatures out of the n keys or addresses.
        2. "keys"       (string, required) A json array of keys which are bitcoin addresses or hex-encoded public keys
             [
               "key"    (string) bitcoin address or hex-encoded public key
               ,...
             ]

        Result:
        {
          "address":"multisigaddress",  (string) The value of the new multisig address.
          "redeemScript":"script"       (string) The string value of the hex-encoded redemption script.
        }
        '''
        return self.call('createmultisig', [nrequired, keys])

    def createrawtransaction(self, inputs, outputs, locktime=0, replaceable=False):
        '''
        http://chainquery.com/bitcoin-api/createrawtransaction

        NOT TESTED

        Create a transaction spending the given inputs and creating new outputs.
        Outputs can be addresses or data.
        Returns hex-encoded raw transaction.
        Note that the transaction's inputs are not signed, and
        it is not stored in the wallet or transmitted to the network.

        Arguments:
        1. "inputs"                (array, required) A json array of json objects
             [
               {
                 "txid":"id",    (string, required) The transaction id
                 "vout":n,         (numeric, required) The output number
                 "sequence":n      (numeric, optional) The sequence number
               }
               ,...
             ]
        2. "outputs"               (object, required) a json object with outputs
            {
              "address": x.xxx,    (numeric or string, required) The key is the bitcoin address, the numeric value (can be string) is the BTC amount
              "data": "hex"      (string, required) The key is "data", the value is hex encoded data
              ,...
            }
        3. locktime                  (numeric, optional, default=0) Raw locktime. Non-0 value also locktime-activates inputs
        4. replaceable               (boolean, optional, default=false) Marks this transaction as BIP125 replaceable.
                                     Allows this transaction to be replaced by a transaction with higher fees. If provided, it is an error if explicit sequence numbers are incompatible.

        Result:
        "transaction"              (string) hex string of the transaction
        '''
        return self.call('createrawtransaction', [inputs, outputs, locktime, replaceable])

    def decoderawtransaction(self, hexstring):
        '''

        TESTED

        http://chainquery.com/bitcoin-api/decoderawtransaction

        Return a JSON object representing the serialized, hex-encoded transaction.

        Arguments:
        1. "hexstring"      (string, required) The transaction hex string

        Result:
        {
          "txid" : "id",        (string) The transaction id
          "hash" : "id",        (string) The transaction hash (differs from txid for witness transactions)
          "size" : n,             (numeric) The transaction size
          "vsize" : n,            (numeric) The virtual transaction size (differs from size for witness transactions)
          "version" : n,          (numeric) The version
          "locktime" : ttt,       (numeric) The lock time
          "vin" : [               (array of json objects)
             {
               "txid": "id",    (string) The transaction id
               "vout": n,         (numeric) The output number
               "scriptSig": {     (json object) The script
                 "asm": "asm",  (string) asm
                 "hex": "hex"   (string) hex
               },
               "txinwitness": ["hex", ...] (array of string) hex-encoded witness data (if any)
               "sequence": n     (numeric) The script sequence number
             }
             ,...
          ],
          "vout" : [             (array of json objects)
             {
               "value" : x.xxx,            (numeric) The value in BTC
               "n" : n,                    (numeric) index
               "scriptPubKey" : {          (json object)
                 "asm" : "asm",          (string) the asm
                 "hex" : "hex",          (string) the hex
                 "reqSigs" : n,            (numeric) The required sigs
                 "type" : "pubkeyhash",  (string) The type, eg 'pubkeyhash'
                 "addresses" : [           (json array of string)
                   "12tvKAXCxZjSmdNbao16dKXC8tRWfcF5oc"   (string) bitcoin address
                   ,...
                 ]
               }
             }
             ,...
          ],
        }
        '''
        return self.call('decoderawtransaction', [hexstring])

    def decodescript(self, hexstring):
        '''

        TESTED

        http://chainquery.com/bitcoin-api/decodescript

        Decode a hex-encoded script.

        Arguments:
        1. "hexstring"     (string) the hex encoded script

        Result:
        {
          "asm":"asm",   (string) Script public key
          "hex":"hex",   (string) hex encoded public key
          "type":"type", (string) The output type
          "reqSigs": n,    (numeric) The required signatures
          "addresses": [   (json array of string)
             "address"     (string) bitcoin address
             ,...
          ],
          "p2sh","address" (string) address of P2SH script wrapping this redeem script (not returned if the script is already a P2SH).
        }
        '''
        return self.call('decodescript', [hexstring])

    def dumpprivkey(self, address):
        '''
        http://chainquery.com/bitcoin-api/dumpprivkey


        Reveals the private key corresponding to 'address'.
        Then the importprivkey can be used with this output

        Arguments:
        1. "address"   (string, required) The bitcoin address for the private key

        Result:
        "key"                (string) The private key
        '''
        return self.call('dumpprivkey', [address])

    def encryptwallet(self, passphrase):
        '''
        http://chainquery.com/bitcoin-api/encryptwallet

        Parameter #1—a passphrase - The passphrase to use for the encrypted wallet. Must be at least one character.
        Result—a notice (with program shutdown on success)

        '''
        return self.call('encryptwallet', [passphrase])

    def estimatefee(self, nblocks):
        '''
        http://chainquery.com/bitcoin-api/estimatefee

        DEPRECATED. Please use estimatesmartfee for more intelligent estimates.
        Estimates the approximate fee per kilobyte needed for a transaction to begin
        confirmation within nblocks blocks. Uses virtual transaction size of transaction
        as defined in BIP 141 (witness data is discounted).

        Arguments:
        1. nblocks     (numeric, required)

        Result:
        n              (numeric) estimated fee-per-kilobyte

        A negative value is returned if not enough transactions and blocks
        have been observed to make an estimate.
        -1 is always returned for nblocks == 1 as it is impossible to calculate
        a fee that is high enough to get reliably included in the next block.
        '''
        return self.call('estimatefee', [nblocks])

    def estimatepriority(self, number):
        '''
        http://chainquery.com/bitcoin-api/estimatepriority

        The estimatepriority RPC estimates the priority that a transaction needs in order to be included within a certain number of blocks as a free high-priority transaction. Transaction priority is relative to a transaction’s byte size.

        Parameter #1—how many blocks the transaction may wait before being included as a free high-priority transaction
        Result—the priority a transaction needs
        '''
        return self.call('estimatepriority', [number])

    def getaccountaddress(self, account):
        '''
        http://chainquery.com/bitcoin-api/getaccountaddress

        DEPRECATED. Returns the current Bitcoin address for receiving payments to this account.

        Arguments:
        1. "account"       (string, required) The account name for the address. It can also be set to the empty string "" to represent the default account. The account does not need to exist, it will be created and a new address created  if there is no account by the given name.

        Result:
        "address"          (string) The account bitcoin address

        '''
        return self.call('getaccountaddress', [account])

    def getaccount(self, address):
        '''
        http://chainquery.com/bitcoin-api/getaccount

        DEPRECATED. Returns the account associated with the given address.

        Arguments:
        1. "address"         (string, required) The bitcoin address for account lookup.

        Result:
        "accountname"        (string) the account address

        '''
        return self.call('getaccount', [address])

    def getaddednodeinfo(self, node):
        '''
        http://chainquery.com/bitcoin-api/getaddednodeinfo

        Returns information about the given added node, or all added nodes
        (note that onetry addnodes are not listed here)

        Arguments:
        1. "node"   (string, optional) If provided, return information about this specific node, otherwise all nodes are returned.

        Result:
        [
          {
            "addednode" : "192.168.0.201",   (string) The node IP address or name (as provided to addnode)
            "connected" : true|false,          (boolean) If connected
            "addresses" : [                    (list of objects) Only when connected = true
               {
                 "address" : "192.168.0.201:8333",  (string) The bitcoin server IP and port we're connected to
                 "connected" : "outbound"           (string) connection, inbound or outbound
               }
             ]
          }
          ,...
        ]

        '''
        return self.call('getaddednodeinfo', [node])

    def getaddressesbyaccount(self, account):
        '''
        http://chainquery.com/bitcoin-api/getaddressesbyaccount

        DEPRECATED. Returns the list of addresses for the given account.

        Arguments:
        1. "account"        (string, required) The account name.

        Result:
        [                     (json array of string)
          "address"         (string) a bitcoin address associated with the given account
          ,...
        ]
        '''
        return self.call('getaddressesbyaccount', [account])

    def getbalance(self, account="*", minconf=1, include_watchonly=False):
        '''
        http://chainquery.com/bitcoin-api/getbalance

        If account is not specified, returns the server's total available balance.
        If account is specified (DEPRECATED), returns the balance in the account.
        Note that the account "" is not the same as leaving the parameter out.
        The server total may be different to the balance in the default "" account.

        Arguments:
        1. "account"         (string, optional) DEPRECATED. The account string may be given as a
                             specific account name to find the balance associated with wallet keys in
                             a named account, or as the empty string ("") to find the balance
                             associated with wallet keys not in any named account, or as "*" to find
                             the balance associated with all wallet keys regardless of account.
                             When this option is specified, it calculates the balance in a different
                             way than when it is not specified, and which can count spends twice when
                             there are conflicting pending transactions (such as those created by
                             the bumpfee command), temporarily resulting in low or even negative
                             balances. In general, account balance calculation is not considered
                             reliable and has resulted in confusing outcomes, so it is recommended to
                             avoid passing this argument.
        2. minconf           (numeric, optional, default=1) Only include transactions confirmed at least this many times.
        3. include_watchonly (bool, optional, default=false) Also include balance in watch-only addresses (see 'importaddress')

        Result:
        amount              (numeric) The total amount in BTC received for this account.

        '''
        return self.call('getbalance', [account, minconf, include_watchonly])

    def getbestblockhash(self):
        '''

        TESTED

        http://chainquery.com/bitcoin-api/getbestblockhash

        Returns the hash of the best (tip) block in the longest blockchain.

        Result:
        "hex"      (string) the block hash hex encoded

        '''
        return self.call('getbestblockhash')

    def getblock(self, blockhash, verbosity=1):
        '''

        TESTED

        http://chainquery.com/bitcoin-api/getblock

        If verbosity is 0, returns a string that is serialized, hex-encoded data for block 'hash'.
        If verbosity is 1, returns an Object with information about block <hash>.
        If verbosity is 2, returns an Object with information about block <hash> and information about each transaction. 

        Arguments:
        1. "blockhash"          (string, required) The block hash
        2. verbosity              (numeric, optional, default=1) 0 for hex encoded data, 1 for a json object, and 2 for json object with transaction data

        Result (for verbosity = 0):
        "data"             (string) A string that is serialized, hex-encoded data for block 'hash'.

        Result (for verbosity = 1):
        {
          "hash" : "hash",     (string) the block hash (same as provided)
          "confirmations" : n,   (numeric) The number of confirmations, or -1 if the block is not on the main chain
          "size" : n,            (numeric) The block size
          "strippedsize" : n,    (numeric) The block size excluding witness data
          "weight" : n           (numeric) The block weight as defined in BIP 141
          "height" : n,          (numeric) The block height or index
          "version" : n,         (numeric) The block version
          "versionHex" : "00000000", (string) The block version formatted in hexadecimal
          "merkleroot" : "xxxx", (string) The merkle root
          "tx" : [               (array of string) The transaction ids
             "transactionid"     (string) The transaction id
             ,...
          ],
          "time" : ttt,          (numeric) The block time in seconds since epoch (Jan 1 1970 GMT)
          "mediantime" : ttt,    (numeric) The median block time in seconds since epoch (Jan 1 1970 GMT)
          "nonce" : n,           (numeric) The nonce
          "bits" : "1d00ffff", (string) The bits
          "difficulty" : x.xxx,  (numeric) The difficulty
          "chainwork" : "xxxx",  (string) Expected number of hashes required to produce the chain up to this block (in hex)
          "previousblockhash" : "hash",  (string) The hash of the previous block
          "nextblockhash" : "hash"       (string) The hash of the next block
        }

        Result (for verbosity = 2):
        {
          ...,                     Same output as verbosity = 1.
          "tx" : [               (array of Objects) The transactions in the format of the getrawtransaction RPC. Different from verbosity = 1 "tx" result.
                 ,...
          ],
          ,...                     Same output as verbosity = 1.
        }
        '''
        return self.call('getblock', [blockhash, verbosity])

    def getblockchaininfo(self):
        '''

        TESTED

        http://chainquery.com/bitcoin-api/getblockchaininfo

        Returns an object containing various state info regarding blockchain processing.

        Result:
        {
          "chain": "xxxx",        (string) current network name as defined in BIP70 (main, test, regtest)
          "blocks": xxxxxx,         (numeric) the current number of blocks processed in the server
          "headers": xxxxxx,        (numeric) the current number of headers we have validated
          "bestblockhash": "...", (string) the hash of the currently best block
          "difficulty": xxxxxx,     (numeric) the current difficulty
          "mediantime": xxxxxx,     (numeric) median time for the current best block
          "verificationprogress": xxxx, (numeric) estimate of verification progress [0..1]
          "chainwork": "xxxx"     (string) total amount of work in active chain, in hexadecimal
          "pruned": xx,             (boolean) if the blocks are subject to pruning
          "pruneheight": xxxxxx,    (numeric) lowest-height complete block stored
          "softforks": [            (array) status of softforks in progress
             {
                "id": "xxxx",        (string) name of softfork
                "version": xx,         (numeric) block version
                "reject": {            (object) progress toward rejecting pre-softfork blocks
                   "status": xx,       (boolean) true if threshold reached
                },
             }, ...
          ],
          "bip9_softforks": {          (object) status of BIP9 softforks in progress
             "xxxx" : {                (string) name of the softfork
                "status": "xxxx",    (string) one of "defined", "started", "locked_in", "active", "failed"
                "bit": xx,             (numeric) the bit (0-28) in the block version field used to signal this softfork (only for "started" status)
                "startTime": xx,       (numeric) the minimum median time past of a block at which the bit gains its meaning
                "timeout": xx,         (numeric) the median time past of a block at which the deployment is considered failed if not yet locked in
                "since": xx,           (numeric) height of the first block to which the status applies
                "statistics": {        (object) numeric statistics about BIP9 signalling for a softfork (only for "started" status)
                   "period": xx,       (numeric) the length in blocks of the BIP9 signalling period 
                   "threshold": xx,    (numeric) the number of blocks with the version bit set required to activate the feature 
                   "elapsed": xx,      (numeric) the number of blocks elapsed since the beginning of the current period 
                   "count": xx,        (numeric) the number of blocks with the version bit set in the current period 
                   "possible": xx      (boolean) returns false if there are not enough blocks left in this period to pass activation threshold 
                }
             }
          }
        }

        '''
        return self.call('getblockchaininfo')

    def getblockcount(self):
        '''

        TESTED

        http://chainquery.com/bitcoin-api/getblockcount

        Returns the number of blocks in the longest blockchain.

        Result:
        n    (numeric) The current block count


        '''
        return self.call('getblockcount')

    def getblockhash(self, height):
        '''

        TESTED

        http://chainquery.com/bitcoin-api/getblockhash

        Returns hash of block in best-block-chain at height provided.

        Arguments:
        1. height         (numeric, required) The height index

        Result:
        "hash"         (string) The block hash

        '''
        return self.call('getblockhash', [height])

    def getblocktemplate(self, template_request=None):
        '''

        TESTED

        http://chainquery.com/bitcoin-api/getblocktemplate

        If the request parameters include a 'mode' key, that is used to explicitly select between the default 'template' request or a 'proposal'.
        It returns data needed to construct a block to work on.
        For full specification, see BIPs 22, 23, 9, and 145:
            https://github.com/bitcoin/bips/blob/master/bip-0022.mediawiki
            https://github.com/bitcoin/bips/blob/master/bip-0023.mediawiki
            https://github.com/bitcoin/bips/blob/master/bip-0009.mediawiki#getblocktemplate_changes
            https://github.com/bitcoin/bips/blob/master/bip-0145.mediawiki

        Arguments:
        1. template_request         (json object, optional) A json object in the following spec
             {
               "mode":"template"    (string, optional) This must be set to "template", "proposal" (see BIP 23), or omitted
               "capabilities":[     (array, optional) A list of strings
                   "support"          (string) client side supported feature, 'longpoll', 'coinbasetxn', 'coinbasevalue', 'proposal', 'serverlist', 'workid'
                   ,...
               ],
               "rules":[            (array, optional) A list of strings
                   "support"          (string) client side supported softfork deployment
                   ,...
               ]
             }


        Result:
        {
          "version" : n,                    (numeric) The preferred block version
          "rules" : [ "rulename", ... ],    (array of strings) specific block rules that are to be enforced
          "vbavailable" : {                 (json object) set of pending, supported versionbit (BIP 9) softfork deployments
              "rulename" : bitnumber          (numeric) identifies the bit number as indicating acceptance and readiness for the named softfork rule
              ,...
          },
          "vbrequired" : n,                 (numeric) bit mask of versionbits the server requires set in submissions
          "previousblockhash" : "xxxx",     (string) The hash of current highest block
          "transactions" : [                (array) contents of non-coinbase transactions that should be included in the next block
              {
                 "data" : "xxxx",             (string) transaction data encoded in hexadecimal (byte-for-byte)
                 "txid" : "xxxx",             (string) transaction id encoded in little-endian hexadecimal
                 "hash" : "xxxx",             (string) hash encoded in little-endian hexadecimal (including witness data)
                 "depends" : [                (array) array of numbers
                     n                          (numeric) transactions before this one (by 1-based index in 'transactions' list) that must be present in the final block if this one is
                     ,...
                 ],
                 "fee": n,                    (numeric) difference in value between transaction inputs and outputs (in Satoshis); for coinbase transactions, this is a negative Number of the total collected block fees (ie, not including the block subsidy); if key is not present, fee is unknown and clients MUST NOT assume there isn't one
                 "sigops" : n,                (numeric) total SigOps cost, as counted for purposes of block limits; if key is not present, sigop cost is unknown and clients MUST NOT assume it is zero
                 "weight" : n,                (numeric) total transaction weight, as counted for purposes of block limits
                 "required" : true|false      (boolean) if provided and true, this transaction must be in the final block
              }
              ,...
          ],
          "coinbaseaux" : {                 (json object) data that should be included in the coinbase's scriptSig content
              "flags" : "xx"                  (string) key name is to be ignored, and value included in scriptSig
          },
          "coinbasevalue" : n,              (numeric) maximum allowable input to coinbase transaction, including the generation award and transaction fees (in Satoshis)
          "coinbasetxn" : { ... },          (json object) information for coinbase transaction
          "target" : "xxxx",                (string) The hash target
          "mintime" : xxx,                  (numeric) The minimum timestamp appropriate for next block time in seconds since epoch (Jan 1 1970 GMT)
          "mutable" : [                     (array of string) list of ways the block template may be changed 
             "value"                          (string) A way the block template may be changed, e.g. 'time', 'transactions', 'prevblock'
             ,...
          ],
          "noncerange" : "00000000ffffffff",(string) A range of valid nonces
          "sigoplimit" : n,                 (numeric) limit of sigops in blocks
          "sizelimit" : n,                  (numeric) limit of block size
          "weightlimit" : n,                (numeric) limit of block weight
          "curtime" : ttt,                  (numeric) current timestamp in seconds since epoch (Jan 1 1970 GMT)
          "bits" : "xxxxxxxx",              (string) compressed target of next block
          "height" : n                      (numeric) The height of the next block
        }

        '''
        if template_request is None:
            return self.call('getblocktemplate', [template_request])
        else:
            return self.call('getblocktemplate')

    def getchaintips(self):
        '''

        TESTED

        http://chainquery.com/bitcoin-api/getchaintips

        Return information about all known tips in the block tree, including the main chain as well as orphaned branches.

        Result:
        [
          {
            "height": xxxx,         (numeric) height of the chain tip
            "hash": "xxxx",         (string) block hash of the tip
            "branchlen": 0          (numeric) zero for main chain
            "status": "active"      (string) "active" for the main chain
          },
          {
            "height": xxxx,
            "hash": "xxxx",
            "branchlen": 1          (numeric) length of branch connecting the tip to the main chain
            "status": "xxxx"        (string) status of the chain (active, valid-fork, valid-headers, headers-only, invalid)
          }
        ]
        Possible values for status:
        1.  "invalid"               This branch contains at least one invalid block
        2.  "headers-only"          Not all blocks for this branch are available, but the headers are valid
        3.  "valid-headers"         All blocks are available for this branch, but they were never fully validated
        4.  "valid-fork"            This branch is not part of the active chain, but is fully validated
        5.  "active"                This is the tip of the active main chain, which is certainly valid

        '''
        return self.call('getchaintips')

    def getconnectioncount(self):
        '''

        TESTED

        http://chainquery.com/bitcoin-api/getconnectioncount


        Returns the number of connections to other nodes.

        Result:
        n          (numeric) The connection count

        '''
        return self.call('getconnectioncount')

    def getdifficulty(self):
        '''

        TESTED

        http://chainquery.com/bitcoin-api/getdifficulty

        Returns the proof-of-work difficulty as a multiple of the minimum difficulty.

        Result:
        n.nnn       (numeric) the proof-of-work difficulty as a multiple of the minimum difficulty.

        '''
        return self.call('getdifficulty')

    def getgenerate(self):
        '''
        http://chainquery.com/bitcoin-api/getgenerate

        The getgenerate RPC returns true if the node is set to generate blocks using its CPU.

        Parameters: none
        Result—whether the server is set to generate blocks

        '''
        return self.call('getgenerate')

    def getinfo(self):
        '''

        TESTED

        http://chainquery.com/bitcoin-api/getinfo

        DEPRECATED. Returns an object containing various state info.

        Result:
        {
          "deprecation-warning": "..." (string) warning that the getinfo command is deprecated and will be removed in 0.16
          "version": xxxxx,           (numeric) the server version
          "protocolversion": xxxxx,   (numeric) the protocol version
          "walletversion": xxxxx,     (numeric) the wallet version
          "balance": xxxxxxx,         (numeric) the total bitcoin balance of the wallet
          "blocks": xxxxxx,           (numeric) the current number of blocks processed in the server
          "timeoffset": xxxxx,        (numeric) the time offset
          "connections": xxxxx,       (numeric) the number of connections
          "proxy": "host:port",       (string, optional) the proxy used by the server
          "difficulty": xxxxxx,       (numeric) the current difficulty
          "testnet": true|false,      (boolean) if the server is using testnet or not
          "keypoololdest": xxxxxx,    (numeric) the timestamp (seconds since Unix epoch) of the oldest pre-generated key in the key pool
          "keypoolsize": xxxx,        (numeric) how many new keys are pre-generated
          "unlocked_until": ttt,      (numeric) the timestamp in seconds since epoch (midnight Jan 1 1970 GMT) that the wallet is unlocked for transfers, or 0 if the wallet is locked
          "paytxfee": x.xxxx,         (numeric) the transaction fee set in BTC/kB
          "relayfee": x.xxxx,         (numeric) minimum relay fee for transactions in BTC/kB
          "errors": "..."             (string) any error messages
        }

        '''
        return self.call('getinfo')

    def getmempoolinfo(self):
        '''

        TESTED

        http://chainquery.com/bitcoin-api/getmempoolinfo

        Returns details on the active state of the TX memory pool.

        Result:
        {
          "size": xxxxx,               (numeric) Current tx count
          "bytes": xxxxx,              (numeric) Sum of all virtual transaction sizes as defined in BIP 141. Differs from actual serialized size because witness data is discounted
          "usage": xxxxx,              (numeric) Total memory usage for the mempool
          "maxmempool": xxxxx,         (numeric) Maximum memory usage for the mempool
          "mempoolminfee": xxxxx       (numeric) Minimum feerate (BTC per KB) for tx to be accepted
        }

        '''
        return self.call('getmempoolinfo')

    def getmininginfo(self):
        '''

        TESTED

        http://chainquery.com/bitcoin-api/getmininginfo

        Returns a json object containing mining-related information.
        Result:
        {
          "blocks": nnn,             (numeric) The current block
          "currentblockweight": nnn, (numeric) The last block weight
          "currentblocktx": nnn,     (numeric) The last block transaction
          "difficulty": xxx.xxxxx    (numeric) The current difficulty
          "errors": "..."            (string) Current errors
          "networkhashps": nnn,      (numeric) The network hashes per second
          "pooledtx": n              (numeric) The size of the mempool
          "chain": "xxxx",           (string) current network name as defined in BIP70 (main, test, regtest)
        }

        '''
        return self.call('getmininginfo')

    def getnettotals(self):
        '''

        TESTED

        http://chainquery.com/bitcoin-api/getnettotals

        Returns information about network traffic, including bytes in, bytes out,
        and current time.

        Result:
        {
          "totalbytesrecv": n,   (numeric) Total bytes received
          "totalbytessent": n,   (numeric) Total bytes sent
          "timemillis": t,       (numeric) Current UNIX time in milliseconds
          "uploadtarget":
          {
            "timeframe": n,                         (numeric) Length of the measuring timeframe in seconds
            "target": n,                            (numeric) Target in bytes
            "target_reached": true|false,           (boolean) True if target is reached
            "serve_historical_blocks": true|false,  (boolean) True if serving historical blocks
            "bytes_left_in_cycle": t,               (numeric) Bytes left in current time cycle
            "time_left_in_cycle": t                 (numeric) Seconds left in current time cycle
          }
        }

        '''
        return self.call('getnettotals')

    def getnetworkhashps(self, nblocks=120, height=-1):
        '''

        TESTED

        http://chainquery.com/bitcoin-api/getnetworkhashps

        Returns the estimated network hashes per second based on the last n blocks.
        Pass in [blocks] to override # of blocks, -1 specifies since last difficulty change.
        Pass in [height] to estimate the network speed at the time when a certain block was found.

        Arguments:
        1. nblocks     (numeric, optional, default=120) The number of blocks, or -1 for blocks since last difficulty change.
        2. height      (numeric, optional, default=-1) To estimate at the time of the given height.

        Result:
        x             (numeric) Hashes per second estimated

        '''
        return self.call('getnetworkhashps', [nblocks, height])

    def getnetworkinfo(self):
        '''

        TESTED

        http://chainquery.com/bitcoin-api/getnetworkinfo

        Returns an object containing various state info regarding P2P networking.

        Result:
        {
          "version": xxxxx,                      (numeric) the server version
          "subversion": "/Satoshi:x.x.x/",     (string) the server subversion string
          "protocolversion": xxxxx,              (numeric) the protocol version
          "localservices": "xxxxxxxxxxxxxxxx", (string) the services we offer to the network
          "localrelay": true|false,              (bool) true if transaction relay is requested from peers
          "timeoffset": xxxxx,                   (numeric) the time offset
          "connections": xxxxx,                  (numeric) the number of connections
          "networkactive": true|false,           (bool) whether p2p networking is enabled
          "networks": [                          (array) information per network
          {
            "name": "xxx",                     (string) network (ipv4, ipv6 or onion)
            "limited": true|false,               (boolean) is the network limited using -onlynet?
            "reachable": true|false,             (boolean) is the network reachable?
            "proxy": "host:port"               (string) the proxy that is used for this network, or empty if none
            "proxy_randomize_credentials": true|false,  (string) Whether randomized credentials are used
          }
          ,...
          ],
          "relayfee": x.xxxxxxxx,                (numeric) minimum relay fee for transactions in BTC/kB
          "incrementalfee": x.xxxxxxxx,          (numeric) minimum fee increment for mempool limiting or BIP 125 replacement in BTC/kB
          "localaddresses": [                    (array) list of local addresses
          {
            "address": "xxxx",                 (string) network address
            "port": xxx,                         (numeric) network port
            "score": xxx                         (numeric) relative score
          }
          ,...
          ]
          "warnings": "..."                    (string) any network warnings
        }

        '''
        return self.call('getnetworkinfo')

    def getpeerinfo(self):
        '''

        TESTED

        http://chainquery.com/bitcoin-api/getpeerinfo

        Returns data about each connected network node as a json array of objects.

        Result:
        [
          {
            "id": n,                   (numeric) Peer index
            "addr":"host:port",      (string) The IP address and port of the peer
            "addrbind":"ip:port",    (string) Bind address of the connection to the peer
            "addrlocal":"ip:port",   (string) Local address as reported by the peer
            "services":"xxxxxxxxxxxxxxxx",   (string) The services offered
            "relaytxes":true|false,    (boolean) Whether peer has asked us to relay transactions to it
            "lastsend": ttt,           (numeric) The time in seconds since epoch (Jan 1 1970 GMT) of the last send
            "lastrecv": ttt,           (numeric) The time in seconds since epoch (Jan 1 1970 GMT) of the last receive
            "bytessent": n,            (numeric) The total bytes sent
            "bytesrecv": n,            (numeric) The total bytes received
            "conntime": ttt,           (numeric) The connection time in seconds since epoch (Jan 1 1970 GMT)
            "timeoffset": ttt,         (numeric) The time offset in seconds
            "pingtime": n,             (numeric) ping time (if available)
            "minping": n,              (numeric) minimum observed ping time (if any at all)
            "pingwait": n,             (numeric) ping wait (if non-zero)
            "version": v,              (numeric) The peer version, such as 7001
            "subver": "/Satoshi:0.8.5/",  (string) The string version
            "inbound": true|false,     (boolean) Inbound (true) or Outbound (false)
            "addnode": true|false,     (boolean) Whether connection was due to addnode and is using an addnode slot
            "startingheight": n,       (numeric) The starting height (block) of the peer
            "banscore": n,             (numeric) The ban score
            "synced_headers": n,       (numeric) The last header we have in common with this peer
            "synced_blocks": n,        (numeric) The last block we have in common with this peer
            "inflight": [
               n,                        (numeric) The heights of blocks we're currently asking from this peer
               ...
            ],
            "whitelisted": true|false, (boolean) Whether the peer is whitelisted
            "bytessent_per_msg": {
               "addr": n,              (numeric) The total bytes sent aggregated by message type
               ...
            },
            "bytesrecv_per_msg": {
               "addr": n,              (numeric) The total bytes received aggregated by message type
               ...
            }
          }
          ,...
        ]

        '''
        return self.call('getpeerinfo')

    def getrawmempool(self, verbose=False):
        '''

        TESTED

        http://chainquery.com/bitcoin-api/getrawmempool

        Returns all transaction ids in memory pool as a json array of string transaction ids.

        Hint: use getmempoolentry to fetch a specific transaction from the mempool.

        Arguments:
        1. verbose (boolean, optional, default=false) True for a json object, false for array of transaction ids

        Result: (for verbose = false):
        [                     (json array of string)
          "transactionid"     (string) The transaction id
          ,...
        ]

        Result: (for verbose = true):
        {                           (json object)
          "transactionid" : {       (json object)
            "size" : n,             (numeric) virtual transaction size as defined in BIP 141. This is different from actual serialized size for witness transactions as witness data is discounted.
            "fee" : n,              (numeric) transaction fee in BTC
            "modifiedfee" : n,      (numeric) transaction fee with fee deltas used for mining priority
            "time" : n,             (numeric) local time transaction entered pool in seconds since 1 Jan 1970 GMT
            "height" : n,           (numeric) block height when transaction entered pool
            "descendantcount" : n,  (numeric) number of in-mempool descendant transactions (including this one)
            "descendantsize" : n,   (numeric) virtual transaction size of in-mempool descendants (including this one)
            "descendantfees" : n,   (numeric) modified fees (see above) of in-mempool descendants (including this one)
            "ancestorcount" : n,    (numeric) number of in-mempool ancestor transactions (including this one)
            "ancestorsize" : n,     (numeric) virtual transaction size of in-mempool ancestors (including this one)
            "ancestorfees" : n,     (numeric) modified fees (see above) of in-mempool ancestors (including this one)
            "depends" : [           (array) unconfirmed transactions used as inputs for this transaction
                "transactionid",    (string) parent transaction id
               ... ]
          }, ...
        }

        '''
        return self.call('getrawmempool', [verbose])

    def getrawtransaction(self, txid, verbose=False):
        '''

        TESTED

        http://chainquery.com/bitcoin-api/getrawtransaction

        NOTE: By default this function only works for mempool transactions. If the -txindex option is
        enabled, it also works for blockchain transactions.
        DEPRECATED: for now, it also works for transactions with unspent outputs.

        Return the raw transaction data.

        If verbose is 'true', returns an Object with information about 'txid'.
        If verbose is 'false' or omitted, returns a string that is serialized, hex-encoded data for 'txid'.

        Arguments:
        1. "txid"      (string, required) The transaction id
        2. verbose       (bool, optional, default=false) If false, return a string, otherwise return a json object

        Result (if verbose is not set or set to false):
        "data"      (string) The serialized, hex-encoded data for 'txid'

        Result (if verbose is set to true):
        {
          "hex" : "data",       (string) The serialized, hex-encoded data for 'txid'
          "txid" : "id",        (string) The transaction id (same as provided)
          "hash" : "id",        (string) The transaction hash (differs from txid for witness transactions)
          "size" : n,             (numeric) The serialized transaction size
          "vsize" : n,            (numeric) The virtual transaction size (differs from size for witness transactions)
          "version" : n,          (numeric) The version
          "locktime" : ttt,       (numeric) The lock time
          "vin" : [               (array of json objects)
             {
               "txid": "id",    (string) The transaction id
               "vout": n,         (numeric)
               "scriptSig": {     (json object) The script
                 "asm": "asm",  (string) asm
                 "hex": "hex"   (string) hex
               },
               "sequence": n      (numeric) The script sequence number
               "txinwitness": ["hex", ...] (array of string) hex-encoded witness data (if any)
             }
             ,...
          ],
          "vout" : [              (array of json objects)
             {
               "value" : x.xxx,            (numeric) The value in BTC
               "n" : n,                    (numeric) index
               "scriptPubKey" : {          (json object)
                 "asm" : "asm",          (string) the asm
                 "hex" : "hex",          (string) the hex
                 "reqSigs" : n,            (numeric) The required sigs
                 "type" : "pubkeyhash",  (string) The type, eg 'pubkeyhash'
                 "addresses" : [           (json array of string)
                   "address"        (string) bitcoin address
                   ,...
                 ]
               }
             }
             ,...
          ],
          "blockhash" : "hash",   (string) the block hash
          "confirmations" : n,      (numeric) The confirmations
          "time" : ttt,             (numeric) The transaction time in seconds since epoch (Jan 1 1970 GMT)
          "blocktime" : ttt         (numeric) The block time in seconds since epoch (Jan 1 1970 GMT)
        }

        '''
        return self.call('getrawtransaction', [txid, verbose])

    def getreceivedbyaccount(self, account, minconf=1):
        '''

        TESTED

        http://chainquery.com/bitcoin-api/getreceivedbyaccount

        DEPRECATED. Returns the total amount received by addresses with <account> in transactions with at least [minconf] confirmations.

        Arguments:
        1. "account"      (string, required) The selected account, may be the default account using "".
        2. minconf          (numeric, optional, default=1) Only include transactions confirmed at least this many times.

        Result:
        amount              (numeric) The total amount in BTC received for this account.

        '''
        return self.call('getreceivedbyaccount', [account, minconf])

    def getreceivedbyaddress(self, address, minconf=1):
        '''

        TESTED

        http://chainquery.com/bitcoin-api/getreceivedbyaddress

        Returns the total amount received by the given address in transactions with at least minconf confirmations.

        Arguments:
        1. "address"         (string, required) The bitcoin address for transactions.
        2. minconf             (numeric, optional, default=1) Only include transactions confirmed at least this many times.

        Result:
        amount   (numeric) The total amount in BTC received at this address.

        '''
        return self.call('getreceivedbyaddress', [address, minconf])

    def gettransaction(self, txid, include_watchonly=False):
        '''
        http://chainquery.com/bitcoin-api/gettransaction

        Get detailed information about in-wallet transaction <txid>

        Arguments:
        1. "txid"                  (string, required) The transaction id
        2. "include_watchonly"     (bool, optional, default=false) Whether to include watch-only addresses in balance calculation and details[]

        Result:
        {
          "amount" : x.xxx,        (numeric) The transaction amount in BTC
          "fee": x.xxx,            (numeric) The amount of the fee in BTC. This is negative and only available for the 
                                      'send' category of transactions.
          "confirmations" : n,     (numeric) The number of confirmations
          "blockhash" : "hash",  (string) The block hash
          "blockindex" : xx,       (numeric) The index of the transaction in the block that includes it
          "blocktime" : ttt,       (numeric) The time in seconds since epoch (1 Jan 1970 GMT)
          "txid" : "transactionid",   (string) The transaction id.
          "time" : ttt,            (numeric) The transaction time in seconds since epoch (1 Jan 1970 GMT)
          "timereceived" : ttt,    (numeric) The time received in seconds since epoch (1 Jan 1970 GMT)
          "bip125-replaceable": "yes|no|unknown",  (string) Whether this transaction could be replaced due to BIP125 (replace-by-fee);
                                                           may be unknown for unconfirmed transactions not in the mempool
          "details" : [
            {
              "account" : "accountname",      (string) DEPRECATED. The account name involved in the transaction, can be "" for the default account.
              "address" : "address",          (string) The bitcoin address involved in the transaction
              "category" : "send|receive",    (string) The category, either 'send' or 'receive'
              "amount" : x.xxx,                 (numeric) The amount in BTC
              "label" : "label",              (string) A comment for the address/transaction, if any
              "vout" : n,                       (numeric) the vout value
              "fee": x.xxx,                     (numeric) The amount of the fee in BTC. This is negative and only available for the 
                                                   'send' category of transactions.
              "abandoned": xxx                  (bool) 'true' if the transaction has been abandoned (inputs are respendable). Only available for the 
                                                   'send' category of transactions.
            }
            ,...
          ],
          "hex" : "data"         (string) Raw data for transaction
        }

        '''
        return self.call('gettransaction', [txid, include_watchonly])

    def gettxout(self, txid, n, include_mempool=True):
        '''

        TESTED

        http://chainquery.com/bitcoin-api/gettxout

        Returns details about an unspent transaction output.

        Arguments:
        1. "txid"             (string, required) The transaction id
        2. "n"                (numeric, required) vout number
        3. "include_mempool"  (boolean, optional) Whether to include the mempool. Default: true.     Note that an unspent output that is spent in the mempool won't appear.

        Result:
        {
          "bestblock" : "hash",    (string) the block hash
          "confirmations" : n,       (numeric) The number of confirmations
          "value" : x.xxx,           (numeric) The transaction value in BTC
          "scriptPubKey" : {         (json object)
             "asm" : "code",       (string)
             "hex" : "hex",        (string)
             "reqSigs" : n,          (numeric) Number of required signatures
             "type" : "pubkeyhash", (string) The type, eg pubkeyhash
             "addresses" : [          (array of string) array of bitcoin addresses
                "address"     (string) bitcoin address
                ,...
             ]
          },
          "coinbase" : true|false   (boolean) Coinbase or not
        }
        '''
        return self.call('gettxout', [txid, n, include_mempool])

    def gettxoutproof(self, txids, blockhash=None):
        '''

        TESTED

        http://chainquery.com/bitcoin-api/gettxoutproof

        Returns a hex-encoded proof that "txid" was included in a block.

        NOTE: By default this function only works sometimes. This is when there is an
        unspent output in the utxo for this transaction. To make it always work,
        you need to maintain a transaction index, using the -txindex command line option or
        specify the block in which the transaction is included manually (by blockhash).

        Arguments:
        1. "txids"       (string) A json array of txids to filter
            [
              "txid"     (string) A transaction hash
              ,...
            ]
        2. "blockhash"   (string, optional) If specified, looks for txid in the block with this hash

        Result:
        "data"           (string) A string that is a serialized, hex-encoded data for the proof.

        '''
        if blockhash is None:
            return self.call('gettxoutproof', [txids])
        else:
            return self.call('gettxoutproof', [txids, blockhash])

    def getunconfirmedbalance(self):
        '''

        TESTED

        http://chainquery.com/bitcoin-api/getunconfirmedbalance

        Returns the server's total unconfirmed balance
        '''
        return self.call('getunconfirmedbalance')

    def getwalletinfo(self):
        '''

        TESTED

        http://chainquery.com/bitcoin-api/getwalletinfo

        Returns an object containing various wallet state info.

        Result:
        {
          "walletname": xxxxx,             (string) the wallet name
          "walletversion": xxxxx,          (numeric) the wallet version
          "balance": xxxxxxx,              (numeric) the total confirmed balance of the wallet in BTC
          "unconfirmed_balance": xxx,      (numeric) the total unconfirmed balance of the wallet in BTC
          "immature_balance": xxxxxx,      (numeric) the total immature balance of the wallet in BTC
          "txcount": xxxxxxx,              (numeric) the total number of transactions in the wallet
          "keypoololdest": xxxxxx,         (numeric) the timestamp (seconds since Unix epoch) of the oldest pre-generated key in the key pool
          "keypoolsize": xxxx,             (numeric) how many new keys are pre-generated (only counts external keys)
          "keypoolsize_hd_internal": xxxx, (numeric) how many new keys are pre-generated for internal use (used for change outputs, only appears if the wallet is using this feature, otherwise external keys are used)
          "unlocked_until": ttt,           (numeric) the timestamp in seconds since epoch (midnight Jan 1 1970 GMT) that the wallet is unlocked for transfers, or 0 if the wallet is locked
          "paytxfee": x.xxxx,              (numeric) the transaction fee configuration, set in BTC/kB
          "hdmasterkeyid": "<hash160>"     (string) the Hash160 of the HD master pubkey
        }

        '''
        return self.call('getwalletinfo')

    def help(self, command=None):
        '''

        TESTED

        http://chainquery.com/bitcoin-api/help

        List all commands, or get help for a specified command.

        Arguments:
        1. "command"     (string, optional) The command to get help on

        Result:
        "text"     (string) The help text

        '''
        if command is None:
            return self.call('help')
        else:
            return self.call('help', [command])

    def keypoolrefill(self, newsize=100):
        '''

        SEEM TO WORK (no RPC error)

        http://chainquery.com/bitcoin-api/keypoolrefill

        Fills the keypool.
        Requires wallet passphrase to be set with walletpassphrase call.

        Arguments
        1. newsize     (numeric, optional, default=100) The new keypool size

        '''
        return self.call('keypoolrefill', [newsize])

    def listaccounts(self, minconf=1, include_watchonly=False):
        '''
        http://chainquery.com/bitcoin-api/listaccounts


        SEEM TO WORK (no RPC error)

        DEPRECATED. Returns Object that has account names as keys, account balances as values.

        Arguments:
        1. minconf             (numeric, optional, default=1) Only include transactions with at least this many confirmations
        2. include_watchonly   (bool, optional, default=false) Include balances in watch-only addresses (see 'importaddress')

        Result:
        {                      (json object where keys are account names, and values are numeric balances
          "account": x.xxx,  (numeric) The property name is the account name, and the value is the total balance for the account.
          ...
        }

        '''
        return self.call('listaccounts', [minconf, include_watchonly])

    def listaddressgroupings(self):
        '''
        http://chainquery.com/bitcoin-api/listaddressgroupings


        SEEM TO WORK (no RPC error)

        Lists groups of addresses which have had their common ownership
        made public by common use as inputs or as the resulting change
        in past transactions

        Result:
        [
          [
            [
              "address",            (string) The bitcoin address
              amount,                 (numeric) The amount in BTC
              "account"             (string, optional) DEPRECATED. The account
            ]
            ,...
          ]
          ,...
        ]
        '''
        return self.call('listaddressgroupings')

    def listlockunspent(self):
        '''
        http://chainquery.com/bitcoin-api/listlockunspent


        SEEM TO WORK (no RPC error)

        Returns list of temporarily unspendable outputs.
        See the lockunspent call to lock and unlock transactions for spending.

        Result:
        [
          {
            "txid" : "transactionid",     (string) The transaction id locked
            "vout" : n                      (numeric) The vout value
          }
          ,...
        ]

        '''
        return self.call('listlockunspent')

    def listreceivedbyaccount(self, minconf=1, include_empty=False, include_watchonly=False):
        '''
        http://chainquery.com/bitcoin-api/listreceivedbyaccount


        SEEM TO WORK (no RPC error)

        DEPRECATED. List balances by account.

        Arguments:
        1. minconf           (numeric, optional, default=1) The minimum number of confirmations before payments are included.
        2. include_empty     (bool, optional, default=false) Whether to include accounts that haven't received any payments.
        3. include_watchonly (bool, optional, default=false) Whether to include watch-only addresses (see 'importaddress').

        Result:
        [
          {
            "involvesWatchonly" : true,   (bool) Only returned if imported addresses were involved in transaction
            "account" : "accountname",  (string) The account name of the receiving account
            "amount" : x.xxx,             (numeric) The total amount received by addresses with this account
            "confirmations" : n,          (numeric) The number of confirmations of the most recent transaction included
            "label" : "label"           (string) A comment for the address/transaction, if any
          }
          ,...
        ]
        '''
        return self.call('listreceivedbyaccount', [minconf, include_empty, include_watchonly])

    def listreceivedbyaddress(self, minconf=1, include_empty=False, include_watchonly=False):
        '''
        http://chainquery.com/bitcoin-api/listreceivedbyaddress


        SEEM TO WORK (no RPC error)

        List balances by receiving address.

        Arguments:
        1. minconf           (numeric, optional, default=1) The minimum number of confirmations before payments are included.
        2. include_empty     (bool, optional, default=false) Whether to include addresses that haven't received any payments.
        3. include_watchonly (bool, optional, default=false) Whether to include watch-only addresses (see 'importaddress').

        Result:
        [
          {
            "involvesWatchonly" : true,        (bool) Only returned if imported addresses were involved in transaction
            "address" : "receivingaddress",  (string) The receiving address
            "account" : "accountname",       (string) DEPRECATED. The account of the receiving address. The default account is "".
            "amount" : x.xxx,                  (numeric) The total amount in BTC received by the address
            "confirmations" : n,               (numeric) The number of confirmations of the most recent transaction included
            "label" : "label",               (string) A comment for the address/transaction, if any
            "txids": [
               n,                                (numeric) The ids of transactions received with the address 
               ...
            ]
          }
          ,...
        ]
        '''
        return self.call('listreceivedbyaddress', [minconf, include_empty, include_watchonly])

    def listtransactions(self, account="*", count=10, skip=0, include_watchonly=False):
        '''
        http://chainquery.com/bitcoin-api/listtransactions


        SEEM TO WORK (no RPC error)

        Returns up to 'count' most recent transactions skipping the first 'from' transactions for account 'account'.

        Arguments:
        1. "account"    (string, optional) DEPRECATED. The account name. Should be "*".
        2. count          (numeric, optional, default=10) The number of transactions to return
        3. skip           (numeric, optional, default=0) The number of transactions to skip
        4. include_watchonly (bool, optional, default=false) Include transactions to watch-only addresses (see 'importaddress')

        Result:
        [
          {
            "account":"accountname",       (string) DEPRECATED. The account name associated with the transaction. 
                                                        It will be "" for the default account.
            "address":"address",    (string) The bitcoin address of the transaction. Not present for 
                                                        move transactions (category = move).
            "category":"send|receive|move", (string) The transaction category. 'move' is a local (off blockchain)
                                                        transaction between accounts, and not associated with an address,
                                                        transaction id or block. 'send' and 'receive' transactions are 
                                                        associated with an address, transaction id and block details
            "amount": x.xxx,          (numeric) The amount in BTC. This is negative for the 'send' category, and for the
                                                 'move' category for moves outbound. It is positive for the 'receive' category,
                                                 and for the 'move' category for inbound funds.
            "label": "label",       (string) A comment for the address/transaction, if any
            "vout": n,                (numeric) the vout value
            "fee": x.xxx,             (numeric) The amount of the fee in BTC. This is negative and only available for the 
                                                 'send' category of transactions.
            "confirmations": n,       (numeric) The number of confirmations for the transaction. Available for 'send' and 
                                                 'receive' category of transactions. Negative confirmations indicate the
                                                 transaction conflicts with the block chain
            "trusted": xxx,           (bool) Whether we consider the outputs of this unconfirmed transaction safe to spend.
            "blockhash": "hashvalue", (string) The block hash containing the transaction. Available for 'send' and 'receive'
                                                  category of transactions.
            "blockindex": n,          (numeric) The index of the transaction in the block that includes it. Available for 'send' and 'receive'
                                                  category of transactions.
            "blocktime": xxx,         (numeric) The block time in seconds since epoch (1 Jan 1970 GMT).
            "txid": "transactionid", (string) The transaction id. Available for 'send' and 'receive' category of transactions.
            "time": xxx,              (numeric) The transaction time in seconds since epoch (midnight Jan 1 1970 GMT).
            "timereceived": xxx,      (numeric) The time received in seconds since epoch (midnight Jan 1 1970 GMT). Available 
                                                  for 'send' and 'receive' category of transactions.
            "comment": "...",       (string) If a comment is associated with the transaction.
            "otheraccount": "accountname",  (string) DEPRECATED. For the 'move' category of transactions, the account the funds came 
                                                  from (for receiving funds, positive amounts), or went to (for sending funds,
                                                  negative amounts).
            "bip125-replaceable": "yes|no|unknown",  (string) Whether this transaction could be replaced due to BIP125 (replace-by-fee);
                                                             may be unknown for unconfirmed transactions not in the mempool
            "abandoned": xxx          (bool) 'true' if the transaction has been abandoned (inputs are respendable). Only available for the 
                                                 'send' category of transactions.
          }
        ]

        '''
        return self.call('listtransactions', [account, count, skip, include_watchonly])

    def listunspent(self, minconf=1, maxconf=9999999, addresses=[], include_unsafe=True, query_options={}):
        '''
        http://chainquery.com/bitcoin-api/listunspent

        SEEM TO WORK (no RPC error)

        Returns array of unspent transaction outputs
        with between minconf and maxconf (inclusive) confirmations.
        Optionally filter to only include txouts paid to specified addresses.

        Arguments:
        1. minconf          (numeric, optional, default=1) The minimum confirmations to filter
        2. maxconf          (numeric, optional, default=9999999) The maximum confirmations to filter
        3. "addresses"      (string) A json array of bitcoin addresses to filter
            [
              "address"     (string) bitcoin address
              ,...
            ]
        4. include_unsafe (bool, optional, default=true) Include outputs that are not safe to spend
                          See description of "safe" attribute below.
        5. query_options    (json, optional) JSON with query options
            {
              "minimumAmount"    (numeric or string, default=0) Minimum value of each UTXO in BTC
              "maximumAmount"    (numeric or string, default=unlimited) Maximum value of each UTXO in BTC
              "maximumCount"     (numeric or string, default=unlimited) Maximum number of UTXOs
              "minimumSumAmount" (numeric or string, default=unlimited) Minimum sum value of all UTXOs in BTC
            }

        Result
        [                   (array of json object)
          {
            "txid" : "txid",          (string) the transaction id
            "vout" : n,               (numeric) the vout value
            "address" : "address",    (string) the bitcoin address
            "account" : "account",    (string) DEPRECATED. The associated account, or "" for the default account
            "scriptPubKey" : "key",   (string) the script key
            "amount" : x.xxx,         (numeric) the transaction output amount in BTC
            "confirmations" : n,      (numeric) The number of confirmations
            "redeemScript" : n        (string) The redeemScript if scriptPubKey is P2SH
            "spendable" : xxx,        (bool) Whether we have the private keys to spend this output
            "solvable" : xxx,         (bool) Whether we know how to spend this output, ignoring the lack of keys
            "safe" : xxx              (bool) Whether this output is considered safe to spend. Unconfirmed transactions
                                      from outside keys and unconfirmed replacement transactions are considered unsafe
                                      and are not eligible for spending by fundrawtransaction and sendtoaddress.
          }
          ,...
        ]
        '''
        return self.call('listunspent', [minconf, maxconf, addresses, include_unsafe, query_options])

    def lockunspent(self, unlock, transactions=None):
        '''

        TESTED

        http://chainquery.com/bitcoin-api/lockunspent

        Updates list of temporarily unspendable outputs.
        Temporarily lock (unlock=false) or unlock (unlock=true) specified transaction outputs.
        If no transaction outputs are specified when unlocking then all current locked transaction outputs are unlocked.
        A locked transaction output will not be chosen by automatic coin selection, when spending bitcoins.
        Locks are stored in memory only. Nodes start with zero locked outputs, and the locked output list
        is always cleared (by virtue of process exit) when a node stops or fails.
        Also see the listunspent call

        Arguments:
        1. unlock            (boolean, required) Whether to unlock (true) or lock (false) the specified transactions
        2. "transactions"  (string, optional) A json array of objects. Each object the txid (string) vout (numeric)
             [           (json array of json objects)
               {
                 "txid":"id",    (string) The transaction id
                 "vout": n         (numeric) The output number
               }
               ,...
             ]

        Result:
        true|false    (boolean) Whether the command was successful or not

        '''
        if transactions is None:
            return self.call('lockunspent', [unlock])
        else:
            return self.call('lockunspent', [unlock, transactions])

    def prioritisetransaction(self, txid, fee_delta):
        '''
        http://chainquery.com/bitcoin-api/prioritisetransaction

        NOT TESTED

        Accepts the transaction into mined blocks at a higher (or lower) priority

        Arguments:
        1. "txid"       (string, required) The transaction id.
        2. dummy          (numeric, optional) API-Compatibility for previous API. Must be zero or null.
                          DEPRECATED. For forward compatibility use named arguments and omit this parameter.
        3. fee_delta      (numeric, required) The fee value (in satoshis) to add (or subtract, if negative).
                          The fee is not actually paid, only the algorithm for selecting transactions into a block
                          considers the transaction as it would have paid a higher (or lower) fee.

        Result:
        true              (boolean) Returns true

        '''
        return self.call('prioritisetransaction', [txid, fee_delta])

    def sendfrom(self, fromaccount, toaddress, amount, minconf=1, comment=None, comment_to=None):
        '''
        http://chainquery.com/bitcoin-api/sendfrom

        NOT TESTED

        DEPRECATED (use sendtoaddress). Sent an amount from an account to a bitcoin address.
        Requires wallet passphrase to be set with walletpassphrase call.

        Arguments:
        1. "fromaccount"       (string, required) The name of the account to send funds from. May be the default account using "".
                               Specifying an account does not influence coin selection, but it does associate the newly created
                               transaction with the account, so the account's balance computation and transaction history can reflect
                               the spend.
        2. "toaddress"         (string, required) The bitcoin address to send funds to.
        3. amount                (numeric or string, required) The amount in BTC (transaction fee is added on top).
        4. minconf               (numeric, optional, default=1) Only use funds with at least this many confirmations.
        5. "comment"           (string, optional) A comment used to store what the transaction is for. 
                                             This is not part of the transaction, just kept in your wallet.
        6. "comment_to"        (string, optional) An optional comment to store the name of the person or organization 
                                             to which you're sending the transaction. This is not part of the transaction, 
                                             it is just kept in your wallet.

        Result:
        "txid"                 (string) The transaction id.

        '''
        param = [fromaccount, toaddress, amount]
        if comment is not None:
            param.append(comment)
        if comment_to is not None:
            param.append(comment_to)
        return self.call('sendfrom', param)

    def sendmany(self, fromaccount, amounts, minconf=1, comment=None, subtractfeefrom=None, replaceable=None, conf_target=None, estimate_mode="UNSET"):
        '''
        http://chainquery.com/bitcoin-api/sendmany

        NOT TESTED

        Send multiple times. Amounts are double-precision floating point numbers.
        Requires wallet passphrase to be set with walletpassphrase call.

        Arguments:
        1. "fromaccount"         (string, required) DEPRECATED. The account to send the funds from. Should be "" for the default account
        2. "amounts"             (string, required) A json object with addresses and amounts
            {
              "address":amount   (numeric or string) The bitcoin address is the key, the numeric amount (can be string) in BTC is the value
              ,...
            }
        3. minconf                 (numeric, optional, default=1) Only use the balance confirmed at least this many times.
        4. "comment"             (string, optional) A comment
        5. subtractfeefrom         (array, optional) A json array with addresses.
                                   The fee will be equally deducted from the amount of each selected address.
                                   Those recipients will receive less bitcoins than you enter in their corresponding amount field.
                                   If no addresses are specified here, the sender pays the fee.
            [
              "address"          (string) Subtract fee from this address
              ,...
            ]
        6. replaceable            (boolean, optional) Allow this transaction to be replaced by a transaction with higher fees via BIP 125
        7. conf_target            (numeric, optional) Confirmation target (in blocks)
        8. "estimate_mode"      (string, optional, default=UNSET) The fee estimate mode, must be one of:
               "UNSET"
               "ECONOMICAL"
               "CONSERVATIVE"

        Result:
        "txid"                   (string) The transaction id for the send. Only 1 transaction is created regardless of 
                                            the number of addresses.

        '''
        param = [fromaccount, amounts, minconf]
        if comment is not None:
            param.append(comment)
        if subtractfeefrom is not None:
            param.append(subtractfeefrom)
        if replaceable is not None:
            param.append(replaceable)
        if conf_target is not None:
            param.append(conf_target)
        param.append(estimate_mode)
        return self.call('sendmany', param)

    def sendrawtransaction(self, hexstring, allowhighfees=False):
        '''
        http://chainquery.com/bitcoin-api/sendrawtransaction

        NOT TESTED

        Submits raw transaction (serialized, hex-encoded) to local node and network.

        Also see createrawtransaction and signrawtransaction calls.

        Arguments:
        1. "hexstring"    (string, required) The hex string of the raw transaction)
        2. allowhighfees    (boolean, optional, default=false) Allow high fees

        Result:
        "hex"             (string) The transaction hash in hex

        '''
        return self.call('sendrawtransaction', [hexstring, allowhighfees])

    def sendtoaddress(self, address, amount, comment=None, comment_to=None, subtractfeefromamount=False, replaceable=None, estimate_mode="UNSET"):
        '''
        http://chainquery.com/bitcoin-api/sendtoaddress

        NOT TESTED

        Send an amount to a given address.

        Requires wallet passphrase to be set with walletpassphrase call.
        Arguments:
        1. "address"            (string, required) The bitcoin address to send to.
        2. "amount"             (numeric or string, required) The amount in BTC to send. eg 0.1
        3. "comment"            (string, optional) A comment used to store what the transaction is for. 
                                     This is not part of the transaction, just kept in your wallet.
        4. "comment_to"         (string, optional) A comment to store the name of the person or organization 
                                     to which you're sending the transaction. This is not part of the 
                                     transaction, just kept in your wallet.
        5. subtractfeefromamount  (boolean, optional, default=false) The fee will be deducted from the amount being sent.
                                     The recipient will receive less bitcoins than you enter in the amount field.
        6. replaceable            (boolean, optional) Allow this transaction to be replaced by a transaction with higher fees via BIP 125
        7. conf_target            (numeric, optional) Confirmation target (in blocks)
        8. "estimate_mode"      (string, optional, default=UNSET) The fee estimate mode, must be one of:
               "UNSET"
               "ECONOMICAL"
               "CONSERVATIVE"

        Result:
        "txid"                  (string) The transaction id.

        '''
        param = [address, amount]
        if comment is not None:
            param.append(comment)
        if comment_to is not None:
            param.append(comment_to)
        param.append(subtractfeefromamount)
        if replaceable is not None:
            param.append(replaceable)
        param.append(estimate_mode)
        return self.call('sendtoaddress', param)

    def settxfee(self, amount):
        '''
        http://chainquery.com/bitcoin-api/settxfee

        SEEM TO WORK (no RPC error)

        Set the transaction fee per kB. Overwrites the paytxfee parameter.

        Arguments:
        1. amount         (numeric or string, required) The transaction fee in BTC/kB

        Result
        true|false        (boolean) Returns true if successful

        '''
        return self.call('settxfee', [amount])

    def signmessage(self, address, message):
        '''
        http://chainquery.com/bitcoin-api/signmessage

        NOT TESTED

        Sign a message with the private key of an address
        Requires wallet passphrase to be set with walletpassphrase call.

        Arguments:
        1. "address"         (string, required) The bitcoin address to use for the private key.
        2. "message"         (string, required) The message to create a signature of.

        Result:
        "signature"          (string) The signature of the message encoded in base 64

        '''
        return self.call('signmessage', [address, message])

    def signrawtransaction(self, hexstring, prevtxs=None, privkeys=None, sighashtype="ALL"):
        '''
        http://chainquery.com/bitcoin-api/signrawtransaction

        NOT TESTED

        Sign inputs for raw transaction (serialized, hex-encoded).
        The second optional argument (may be null) is an array of previous transaction outputs that
        this transaction depends on but may not yet be in the block chain.
        The third optional argument (may be null) is an array of base58-encoded private
        keys that, if given, will be the only keys used to sign the transaction.

        Requires wallet passphrase to be set with walletpassphrase call.

        Arguments:
        1. "hexstring"     (string, required) The transaction hex string
        2. "prevtxs"       (string, optional) An json array of previous dependent transaction outputs
             [               (json array of json objects, or 'null' if none provided)
               {
                 "txid":"id",             (string, required) The transaction id
                 "vout":n,                  (numeric, required) The output number
                 "scriptPubKey": "hex",   (string, required) script key
                 "redeemScript": "hex",   (string, required for P2SH or P2WSH) redeem script
                 "amount": value            (numeric, required) The amount spent
               }
               ,...
            ]
        3. "privkeys"     (string, optional) A json array of base58-encoded private keys for signing
            [                  (json array of strings, or 'null' if none provided)
              "privatekey"   (string) private key in base58-encoding
              ,...
            ]
        4. "sighashtype"     (string, optional, default=ALL) The signature hash type. Must be one of
               "ALL"
               "NONE"
               "SINGLE"
               "ALL|ANYONECANPAY"
               "NONE|ANYONECANPAY"
               "SINGLE|ANYONECANPAY"

        Result:
        {
          "hex" : "value",           (string) The hex-encoded raw transaction with signature(s)
          "complete" : true|false,   (boolean) If the transaction has a complete set of signatures
          "errors" : [                 (json array of objects) Script verification errors (if there are any)
            {
              "txid" : "hash",           (string) The hash of the referenced, previous transaction
              "vout" : n,                (numeric) The index of the output to spent and used as input
              "scriptSig" : "hex",       (string) The hex-encoded signature script
              "sequence" : n,            (numeric) Script sequence number
              "error" : "text"           (string) Verification or signing error related to the input
            }
            ,...
          ]
        }

        '''
        param = [hexstring]
        if prevtxs is not None:
            param.append(prevtxs)
        if privkeys is not None:
            param.append(privkeys)
        param.append(sighashtype)
        return self.call('signrawtransaction', param)

    def submitblock(self, hexdata, dummy=None):
        '''
        http://chainquery.com/bitcoin-api/submitblock

        NOT TESTED

        Attempts to submit new block to network.
        See https://en.bitcoin.it/wiki/BIP_0022 for full specification.

        Arguments
        1. "hexdata"        (string, required) the hex-encoded block data to submit
        2. "dummy"          (optional) dummy value, for compatibility with BIP22. This value is ignored.

        Result: None

        '''
        if dummy is None:
            return self.call('submitblock', [hexdata])
        else:
            return self.call('submitblock', [hexdata, dummy])

    def validateaddress(self, address):
        '''
        http://chainquery.com/bitcoin-api/validateaddress

        TESTED

        Return information about the given bitcoin address.

        Arguments:
        1. "address"     (string, required) The bitcoin address to validate

        Result:
        {
          "isvalid" : true|false,       (boolean) If the address is valid or not. If not, this is the only property returned.
          "address" : "address", (string) The bitcoin address validated
          "scriptPubKey" : "hex",       (string) The hex encoded scriptPubKey generated by the address
          "ismine" : true|false,        (boolean) If the address is yours or not
          "iswatchonly" : true|false,   (boolean) If the address is watchonly
          "isscript" : true|false,      (boolean) If the key is a script
          "script" : "type"             (string, optional) The output script type. Possible types: nonstandard, pubkey, pubkeyhash, scripthash, multisig, nulldata, witness_v0_keyhash, witness_v0_scripthash
          "hex" : "hex",                (string, optional) The redeemscript for the p2sh address
          "addresses"                   (string, optional) Array of addresses associated with the known redeemscript
            [
              "address"
              ,...
            ]
          "sigsrequired" : xxxxx        (numeric, optional) Number of signatures required to spend multisig output
          "pubkey" : "publickeyhex",    (string) The hex value of the raw public key
          "iscompressed" : true|false,  (boolean) If the address is compressed
          "account" : "account"         (string) DEPRECATED. The account associated with the address, "" is the default account
          "timestamp" : timestamp,        (number, optional) The creation time of the key if available in seconds since epoch (Jan 1 1970 GMT)
          "hdkeypath" : "keypath"       (string, optional) The HD keypath if the key is HD and available
          "hdmasterkeyid" : "<hash160>" (string, optional) The Hash160 of the HD master pubkey
        }

        '''
        return self.call('validateaddress', [address])

    def verifymessage(self, address, signature, message):
        '''
        http://chainquery.com/bitcoin-api/verifymessage

        NOT TESTED

        Verify a signed message

        Arguments:
        1. "address"         (string, required) The bitcoin address to use for the signature.
        2. "signature"       (string, required) The signature provided by the signer in base 64 encoding (see signmessage).
        3. "message"         (string, required) The message that was signed.

        Result:
        true|false   (boolean) If the signature is verified or not.

        '''
        return self.call('verifymessage', [address, signature, message])

    def verifytxoutproof(self, proof):
        '''
        http://chainquery.com/bitcoin-api/verifytxoutproof

        TESTED

        Verifies that a proof points to a transaction in a block, returning the transaction it commits to
        and throwing an RPC error if the block is not in our best chain

        Arguments:
        1. "proof"    (string, required) The hex-encoded proof generated by gettxoutproof

        Result:
        ["txid"]      (array, strings) The txid(s) which the proof commits to, or empty array if the proof is invalid

        '''
        return self.call('verifytxoutproof', [proof])

    def walletlock(self):
        '''
        http://chainquery.com/bitcoin-api/walletlock

        NOT TESTED

        Parameters: none
        Result—null on success

        Removes the wallet encryption key from memory, locking the wallet.
        After calling this method, you will need to call walletpassphrase again
        before being able to call any methods which require the wallet to be unlocked.

        '''
        return self.call('walletlock')

    def walletpassphrase(self, passphrase, timeout):
        '''
        http://chainquery.com/bitcoin-api/walletpassphrase

        NOT TESTED

        Stores the wallet decryption key in memory for 'timeout' seconds.
        This is needed prior to performing transactions related to private keys such as sending bitcoins

        Arguments:
        1. "passphrase"     (string, required) The wallet passphrase
        2. timeout            (numeric, required) The time to keep the decryption key in seconds.

        Note:
        Issuing the walletpassphrase command while the wallet is already unlocked will set a new unlock
        time that overrides the old one.

        '''
        return self.call('walletpassphrase', [passphrase, timeout])

    def walletpassphrasechange(self, oldpassphrase, newpassphrase):
        '''
        http://chainquery.com/bitcoin-api/walletpassphrasechange

        NOT TESTED

        Changes the wallet passphrase from 'oldpassphrase' to 'newpassphrase'.

        Arguments:
        1. "oldpassphrase"      (string) The current passphrase
        2. "newpassphrase"      (string) The new passphrase

        '''
        return self.call('walletpassphrasechange', [oldpassphrase, newpassphrase])


class BitcoinBitcoreExplorerRPC(BitcoinExplorerRPC):
    '''
    BitcoinExplorerRPC subclass for bitcored-specific methods

    new RPC calls : https://bitcore.io/guides/bitcoin
    '''

    def __init__(self, host):
        BitcoinExplorerRPC.__init__(self, host=host)

    def getaddresstxids(self):
        raise NotImplementedError

    def getblock(self, blockhash, verbosity=True):
        return self.call('getblock', [blockhash, verbosity])

    def __script_pub_key_to_bech32(self, script):
        # remove OP_0 + len(pubkeyhash)
        pubkeyhash = script[4:]
        if script[2:4] == '14':
            type_tx = 'witness_v0_keyhash'
        elif script[2:4] == '20':
            type_tx = 'witness_v0_scripthash'
        else:
            raise Exception('script = %s', script)

        # bc for mainnet
        # 0x00 for the version = v0
        # pubkeyhash
        # return bech32 address
        return bech32_encode('bc', 0x00,
                             binascii.unhexlify(pubkeyhash)), type_tx

    # case TX_NONSTANDARD: return "nonstandard";
    # case TX_PUBKEY: return "pubkey";
    # case TX_PUBKEYHASH: return "pubkeyhash";
    # case TX_SCRIPTHASH: return "scripthash";
    # case TX_MULTISIG: return "multisig";
    # case TX_NULL_DATA: return "nulldata";
    # case TX_WITNESS_V0_KEYHASH: return "witness_v0_keyhash";
    # case TX_WITNESS_V0_SCRIPTHASH: return "witness_v0_scripthash";
    # case TX_WITNESS_UNKNOWN: return "witness_unknown";

    # decode_tx to use extra fields add by bitcore in new getrawtransaction
    def decode_tx(self, tx):

        result = {"txid": tx['txid'],
                  "sender": [],
                  "receiver": [],
                  "total_value_input": 0.0,
                  "total_value_output": 0.0,
                  "fee": 0.0
                  }

        # get the senders
        for vin in tx['vin']:

            # test for commons transaction first
            # vin['address'] & vin['addr'] are bitcore specific
            if 'address' in vin:
                sender_address = vin['address']
                type_tx = 'standard'
            elif 'addr' in vin:
                sender_address = vin['addr']
                type_tx = 'standard'

            # coinbase transaction
            # only one per block
            # first transaction that's the coinbase mining reward
            elif 'coinbase' in vin:
                result['sender'].append(('mining_reward', 0,
                                         '', 'mining_reward'))
                break

            # SegWit or less classic case
            # need to retrieve complete input transaction
            else:
                sender_tx = self.getrawtransaction(vin['txid'], 1)
                for sender_vout in sender_tx['vout']:
                    if sender_vout['n'] == vin['vout']:
                        script_pub_key = sender_vout['scriptPubKey']
                        script_sig = script_pub_key['hex']
                        # pubkey
                        try:
                            sender_address = script_pub_key['addresses'][0]
                            type_tx = script_pub_key['type']
                            print('sender type_tx: %s' % type_tx)
                        # SegWit
                        except KeyError:
                            sender_address, type_tx = \
                                self.__script_pub_key_to_bech32(script_sig)
                            if sender_address is None:
                                raise Exception('sender bech32 : %s',
                                                script_sig)

            sender_value = vin['value']
            sender_script = vin['scriptSig']['hex']
            result['sender'].append((sender_address, sender_value,
                                     sender_script, type_tx))

        # get the receivers
        for vout in tx['vout']:
            value = vout['value']
            script_pub_key = vout['scriptPubKey']
            btcscript = script_pub_key['hex']

            try:
                receiver = script_pub_key['addresses'][0]
                type_tx = script_pub_key['type']
            # KeyError appears when vin['address'] doesn't exit
            # means we have SegWit transaction
            except KeyError:
                if 'type' in script_pub_key:
                    type_tx = script_pub_key['type']
                    if type_tx == 'nulldata':
                        receiver = 'lost_bitcoin'
                    else:
                        #print('receiver type_tx: %s' % type_tx)
                        receiver, type_tx = \
                            self.__script_pub_key_to_bech32(btcscript)
                        if receiver is None:
                            raise Exception('receiver bech32 : %s', script_sig)

            result['receiver'].append((receiver, value, btcscript, type_tx))

        total_in = sum(v for _, v, _, _ in result['sender'])
        total_out = sum(v for _, v, _, _ in result['receiver'])

        for s, v, _, _ in result['sender']:
            if s == '':
                print('TXID = %s' % tx['txid'])

        for r, v, _, _ in result['receiver']:
            if r == '':
                print('TXID = %s' % tx['txid'])

        result["total_value_input"] = total_in
        result["total_value_output"] = total_out
        result["fee"] = total_in - total_out if total_in - total_out > 0 else 0
        return result

# example address p2wsh
# txid = a388d6f88373aacf2a5c170eeff4adf51f2c6744da26375d5890f42d653a2e6c
# txid2 = 37138d3df8e1d41bf6474ecdc21d53f78ab9907c500172de31bda4048556bd83
# pubkeyhash = 701a8d401c84fb13e6baf169d59684e17abd9fa216c8cc5b9fc63d622ff8c58
# address = bc1qwqdg6squsna38e46795at95yu9atm8azzmyvckulcc7kytlcckxswvvzej

# example address p2wpkh
# txid = fa3c4931c3f31eda18df255d754aa8f36a514f6ef7408d963912c47bb5cc5ec7
# txid2 = 761619410d9f2a7fa69336583a0351dd86caf9ed2cee8765132a93caa44f77de
# pubkeyhash = f2c78dda8d18a90545045942989406bf90966f7f
# address = bc1q7trcmk5drz5s23gyt9pf39qxh7gfvmmlr8kc2n

# can cause crash
# 0fcb06da0fb13e112fe68f3d674dc1366ec2930b4852ad8581b8ef7cc5f25fdf
# a562429d601b5559b9e1d732d7ad7c6071243f9b260af3f480757353841321ed

# first tx SegWit
# txid = f91d0a8a78462bc59398f2c5d7a84fcff491c26ba54c4833478b202796c8aafd
# block_number = 481824



from octopus.arch.evm.cfg import EvmCFG
from octopus.arch.wasm.cfg import WasmCFG


# Ethereum smart contract == EVM bytecode or WebAssembly
class EthereumCFG(object):
    def __new__(cls, bytecode, arch='evm', evm_analysis='dynamic'):
        if arch == 'evm':
            return EvmCFG(bytecode, analysis=evm_analysis)
        else:  # Wasm
            return WasmCFG(bytecode)



from wasm.modtypes import (TypeSection,
                           ImportSection,
                           FunctionSection,
                           TableSection,
                           MemorySection,
                           GlobalSection,
                           ExportSection,
                           StartSection,
                           ElementSection,
                           CodeSection,
                           DataSection)

from octopus.arch.wasm.constant import LANG_TYPE, KIND_TYPE
from octopus.arch.wasm.format import (format_kind_function,
                                      format_kind_table,
                                      format_kind_memory,
                                      format_kind_global)

from octopus.arch.wasm.decode import decode_module

from octopus.core.utils import bytecode_to_bytes
# from wasm.decode import decode_module
import io
import json
import os

from logging import getLogger
logging = getLogger(__name__)


class WasmModuleAnalyzer(object):
    '''Analyze and extract informations from wasm module'''

    def __init__(self, module_bytecode, analysis=True):
        self.module_bytecode = bytecode_to_bytes(module_bytecode)

        self.magic = None
        self.version = None
        self.types = list()
        self.imports_all = list()
        self.imports_func = list()
        self.func_types = list()
        self.tables = list()
        self.memories = list()
        self.globals = list()
        self.exports = list()
        self.start = None
        self.elements = list()
        self.codes = list()
        self.datas = list()
        self.names = list()
        self.customs = list()
        self.func_prototypes = list()
        # self.strings = list() - TODO

        if analysis:
            self.analyze()

    def attributes_reset(self):
        self.magic = None
        self.version = None
        self.types = list()
        self.imports_all = list()
        self.imports_func = list()
        self.func_types = list()
        self.tables = list()
        self.memories = list()
        self.globals = list()
        self.exports = list()
        self.start = None
        self.elements = list()
        self.codes = list()
        self.datas = list()
        self.names = list()
        self.customs = list()
        self.func_prototypes = list()

    def __str__(self):
        return str(self.show())

    def show(self):
        """Return dict with WasmModuleAnalyzer attributes"""
        return {'magic': self.magic,
                'version': self.version,
                'types': self.types,
                'imports_all': self.imports_all,
                'imports_func': self.imports_func,
                'func_types': self.func_types,
                'tables': self.tables,
                'memories': self.memories,
                'globals': self.globals,
                'exports': self.exports,
                'start': self.start,
                'elements': self.elements,
                'length codes': len(self.codes),
                'datas': self.datas,
                'func_prototypes': self.func_prototypes}

    def __get_section(self, section_type):
        mod_iter = iter(decode_module(self.module_bytecode))
        _, _ = next(mod_iter)
        sections = list(mod_iter)

        # iterate over all section
        for cur_sec, cur_sec_data in sections:
            sec = cur_sec_data.get_decoder_meta()['types']['payload']

            if isinstance(sec, section_type):
                return cur_sec_data
        return None

    def __decode_header(self, header, h_data):
        """Decode wasm header
        Return tuple (magic, version) of wasm module header
        .. seealso:: https://github.com/WebAssembly/design/blob/master/BinaryEncoding.md#high-level-structure
        """
        magic = \
            h_data.magic.to_bytes(header.magic.byte_size, 'little')
        version = \
            h_data.version.to_bytes(header.version.byte_size, 'little')
        return (magic, version)

    def __decode_type_section(self, type_section):
        """Decode wasm type section
        Return a list of tuple (param_str, return_str)
        .. seealso:: https://github.com/WebAssembly/design/blob/master/BinaryEncoding.md#type-section
        """
        type_list = []

        for idx, entry in enumerate(type_section.payload.entries):
            param_str = ''
            return_str = ''

            param_str += ' '.join([LANG_TYPE.get(_x) for _x in entry.param_types])
            if entry.return_type:
                return_str = '%s' % LANG_TYPE.get(entry.return_type)

            type_list.append((param_str, return_str))
        return type_list

    def __decode_import_section(self, import_section):
        """Decode import section to tuple of list
        .. seealso:: https://github.com/WebAssembly/design/blob/master/BinaryEncoding.md#import-section
        """
        entries = import_section.payload.entries
        import_list = []
        import_func_list = []

        for idx, entry in enumerate(entries):
            #            for encoding in ('utf-8', 'utf-16-be'):
            #                value = str(v)
            #                try:
            #                    value = v.decode(encoding)
            #                    break
            #                except UnicodeDecodeError:
            #                    value = str(v)
            try:
                module_str = entry.module_str.tobytes().decode('utf-8')
            except UnicodeDecodeError:
                module_str = entry.module_str.tobytes()
            try:
                field_str = entry.field_str.tobytes().decode('utf-8')
            except UnicodeDecodeError:
                field_str = entry.field_str.tobytes()

            logging.debug('%s %s', module_str, field_str)
            kind_type = KIND_TYPE.get(entry.kind)

            if kind_type == 'function':
                f_type = format_kind_function(entry.type.type)
                import_list.append((entry.kind, module_str, field_str, f_type))
                # add also the info into the specific import function list
                import_func_list.append((module_str, field_str, f_type))
            elif kind_type == 'table':
                tabl = format_kind_table(entry.type.element_type,
                                         entry.type.limits.flags,
                                         entry.type.limits.initial,
                                         entry.type.limits.maximum)
                import_list.append((entry.kind, module_str, field_str, tabl))

            elif kind_type == 'memory':
                mem = format_kind_memory(entry.type.limits.flags,
                                         entry.type.limits.initial,
                                         entry.type.limits.maximum)
                import_list.append((entry.kind, module_str, field_str, mem))

            elif kind_type == 'global':
                gbl = format_kind_global(entry.type.content_type,
                                         entry.type.mutability)
                import_list.append((entry.kind, module_str, field_str, gbl))
            else:
                logging.error('unknown %d %s %s', entry.kind,
                              module_str, field_str)
        return (import_list, import_func_list)

    def __decode_function_section(self, function_section):
        """Decode function section
        The function section declares the signatures of all functions in the module
        Return list of indices (int)
        .. seealso:: https://github.com/WebAssembly/design/blob/master/BinaryEncoding.md#function-section
        """
        return function_section.payload.types

    def __decode_table_section(self, table_section):
        """
        .. seealso:: https://github.com/WebAssembly/design/blob/master/BinaryEncoding.md#table-section
        """
        # on the MVP, table size == 1
        entries = table_section.payload.entries
        table_list = []

        for idx, entry in enumerate(entries):
            element_type = entry.element_type
            flags = entry.limits.flags
            initial = entry.limits.initial
            maximum = entry.limits.maximum

            fmt = format_kind_table(element_type,
                                    flags,
                                    initial,
                                    maximum)
            table_list.append(fmt)
        return table_list

    def __decode_memory_section(self, memory_section):
        """
        .. seealso:: https://github.com/WebAssembly/design/blob/master/BinaryEncoding.md#memory-section
        """
        # on the MVP, memory size == 1
        memory_l = list()
        entries = memory_section.payload.entries

        for idx, entry in enumerate(entries):

            flags = entry.limits.flags
            initial = entry.limits.initial
            maximum = entry.limits.maximum

            fmt = format_kind_memory(flags,
                                     initial,
                                     maximum)

            memory_l.append(fmt)
        return memory_l

    def __decode_global_section(self, global_section):
        """
        .. seealso:: https://github.com/WebAssembly/design/blob/master/BinaryEncoding.md#global-section
        """
        globals_l = list()
        for entry in global_section.payload.globals:
            fmt = format_kind_global(entry.type.mutability,
                                     entry.type.content_type)
            globals_l.append(fmt)
        return globals_l

    def __decode_export_section(self, export_section):
        """
        .. seealso:: https://github.com/WebAssembly/design/blob/master/BinaryEncoding.md#export-section
        """
        entries = export_section.payload.entries
        export_list = []

        for idx, entry in enumerate(entries):
            # field_str == function_name
            try:
                field_str = entry.field_str.tobytes().decode('utf-8')
            except UnicodeDecodeError:
                field_str = entry.field_str.tobytes()
            kind = entry.kind
            index = entry.index

            fmt = {'field_str': field_str,
                   'kind': kind,
                   'index': index}
            export_list.append(fmt)
        return export_list

    def __decode_start_section(self, start_section):
        return start_section.payload.index

    def __decode_element_section(self, element_section):
        """
        .. seealso:: https://github.com/WebAssembly/design/blob/master/BinaryEncoding.md#element-section
        """
        entries = element_section.payload.entries
        element_list = []

        for idx, entry in enumerate(entries):

            fmt = {'index': entry.index,
                   'offset': entry.offset,
                   'elems': entry.elems}

            element_list.append(fmt)
        return element_list

    def __decode_code_section(self, code_section):
        """
        .. seealso:: https://github.com/WebAssembly/design/blob/master/BinaryEncoding.md#code-section
        """
        bodies = code_section.payload.bodies
        code_list = []

        for idx, entry in enumerate(bodies):
            code_raw = entry.code.tobytes()
            code_list.append(code_raw)
        return code_list

    def __decode_data_section(self, data_section):
        """
        .. seealso:: https://github.com/WebAssembly/design/blob/master/BinaryEncoding.md#data-section
        """
        entries = data_section.payload.entries
        data_list = []

        for idx, entry in enumerate(entries):
            data = entry.data.tobytes()

            fmt = {'index': entry.index,
                   'offset': entry.offset,
                   'size': entry.size,
                   'data': data}

            data_list.append(fmt)
        return data_list

    def __decode_name_section(self, name_section):
        """
        .. seealso:: https://github.com/WebAssembly/design/blob/master/BinaryEncoding.md#name-section
        """
        payload = name_section.payload.tobytes()
        #print(payload)

        total = 0
        names_list = list()

        f = io.BytesIO(payload)
        f.read(2)  # drop 2 bytes
        total += 2
        count = int.from_bytes(f.read(1), byteorder='little')
        total += 1

        while total < len(payload):
            index = int.from_bytes(f.read(1), byteorder='big')
            total += 1
            name_len = int.from_bytes(f.read(1), byteorder='big')
            total += 1
            name_str = f.read(name_len)
            total += name_len
            names_list.append((index, name_len, name_str))
        f.close()
        return names_list

    def __decode_unknown_section(self, unknown_section):
        """
        .. seealso:: https://github.com/WebAssembly/design/blob/master/BinaryEncoding.md#high-level-structure
        """
        sec_name = unknown_section.name.tobytes()
        payload = unknown_section.payload.tobytes()
        return (sec_name, payload)

    def get_func_prototypes_ordered(self):
        """create ordered list of functions"""

        func_prototypes = list()

        # get imported functions
        for _, name, type_idx in self.imports_func:
            _param, _return = self.types[type_idx]
            func_prototypes.append((name, _param, _return, 'import'))

        # get all internal functions
        for idx, code in enumerate(self.codes):
            _param, _return = self.types[self.func_types[idx]]
            real_index = len(self.imports_func) + idx
            name = '$func%d' % real_index
            f_type = 'local'

            # if exported function - overwrite name
            for x in self.exports:
                if x.get('index') == real_index and x.get('kind') == 0:
                    name = x.get('field_str')
                    f_type = 'export'

            # TODO: need to test
            if real_index == self.start:
                name = '* ' + name
            func_prototypes.append((name, _param, _return, f_type))
        return func_prototypes

    def analyze(self):
        """analyse the complete module & extract informations """
        # src: https://github.com/WebAssembly/design/blob/master/BinaryEncoding.md
        # custom     0   name, .debug_str, ...
        # Type       1   Function signature declarations
        # Import     2   Import declarations
        # Function   3   Function declarations
        # Table      4   Indirect function table and other tables
        # Memory     5   Memory attributes
        # Global     6   Global declarations
        # Export     7   Exports
        # Start      8   Start function declaration
        # Element    9   Elements section
        # Code       10  Function bodies (code)
        # Data       11  Data segments

        # reset attributes
        self.attributes_reset()

        mod_iter = iter(decode_module(self.module_bytecode))
        # decode header version - usefull in the future (multiple versions)
        header, header_data = next(mod_iter)
        self.magic, self.version = self.__decode_header(header, header_data)

        #
        # Wasm sections
        #
        sections = list(mod_iter)

        for cur_sec, cur_sec_data in sections:
            sec = cur_sec_data.get_decoder_meta()['types']['payload']

            if isinstance(sec, TypeSection):
                self.types = self.__decode_type_section(cur_sec_data)
            elif isinstance(sec, ImportSection):
                self.imports_all, self.imports_func = \
                    self.__decode_import_section(cur_sec_data)
            elif isinstance(sec, FunctionSection):
                self.func_types = self.__decode_function_section(cur_sec_data)
            elif isinstance(sec, TableSection):
                self.tables = self.__decode_table_section(cur_sec_data)
            elif isinstance(sec, MemorySection):
                self.memories = self.__decode_memory_section(cur_sec_data)
            elif isinstance(sec, GlobalSection):
                # TODO not analyzed
                self.globals = self.__decode_global_section(cur_sec_data)
            elif isinstance(sec, ExportSection):
                self.exports = self.__decode_export_section(cur_sec_data)
            elif isinstance(sec, StartSection):
                # TODO not analyzed
                self.start = self.__decode_start_section(cur_sec_data)
            elif isinstance(sec, ElementSection):
                self.elements = self.__decode_element_section(cur_sec_data)
            elif isinstance(sec, CodeSection):
                self.codes = self.__decode_code_section(cur_sec_data)
            elif isinstance(sec, DataSection):
                self.datas = self.__decode_data_section(cur_sec_data)
            else:
                # name section
                if cur_sec_data.id == 0 and cur_sec_data.name.tobytes() == b'name':
                    self.names = self.__decode_name_section(cur_sec_data)
                else:
                    # TODO - handle properly .debug_str section
                    self.customs.append(self.__decode_unknown_section(cur_sec_data))

        # create ordered list of functions
        self.func_prototypes = self.get_func_prototypes_ordered()
        return True

    def is_compiled_with_emscripten(self):
        matching_list = self.get_emscripten_calls()
        return True if matching_list else False

    def get_emscripten_calls(self):
        res = [x for x, _, _, _ in self.func_prototypes if is_emscripten_func(x)]
        return res

    # emscripten syscall from:
    # * https://github.com/kripken/emscripten/blob/incoming/system/lib/fetch/asmfs.cpp
    # * http://gauss.ececs.uc.edu/Courses/c4029/code/syscall_32.tbl.html
    def contains_emscripten_syscalls(self):
        EMSCRIPTEN_SYSCALL_JSON = '/signatures/emscripten_syscalls.json'
        path = os.path.dirname(os.path.realpath(__file__)) + EMSCRIPTEN_SYSCALL_JSON

        json_data = open(path).read()
        data = json.loads(json_data)

        func_names = [x for x, _, _, _ in self.func_prototypes]
        match = list()
        for name in func_names:
            try:
                # remove '_' to match '__syscallXX' & '___syscallXX'
                syscall = data[name.replace('_', '')]
                match.append((name, syscall))
            except KeyError:
                pass
        return match


def is_emscripten_func(x):

    # from https://github.com/kripken/emscripten/blob/master/emscripten.py
    EMSCRIPTEN_LIST = [
        # create_basic_funcs
        'abort', 'assert', 'enlargeMemory', 'getTotalMemory',
        'abortOnCannotGrowMemory',
        'abortStackOverflow',
        'abortStackOverflowEmterpreter',
        'segfault', 'alignfault', 'ftfault',
        'SAFE_HEAP_LOAD', 'SAFE_HEAP_LOAD_D',
        'SAFE_HEAP_STORE', 'SAFE_HEAP_STORE_D', 'SAFE_FT_MASK',
        # create_receiving
        '_memcpy', '_memset', 'runPostSets',
        '_emscripten_replace_memory', '__start_module',
        # create_asm_runtime_funcs
        'stackAlloc', 'stackSave',
        'stackRestore', 'establishStackSpace', 'setThrew',
        'setTempRet0', 'getTempRet0',
        'setDynamicTop',
        'emterpret',
        'setAsyncState', 'emtStackSave',
        'emtStackRestore', 'getEmtStackMax', 'setEmtStackMax',
        'setAsync',
        'addOnExit',
        # from emscripten/emcc.py
        '_stbi_load', '_stbi_load_from_memory', '_stbi_image_free',
        '___cxa_demangle',
        '_malloc', '_free',
        'allocate', 'getMemory',
        '___errno_location', '_fflush',
        'FS_createFolder',
        'FS_createPath',
        'FS_createDataFile',
        'FS_createPreloadedFile',
        'FS_createLazyFile',
        'FS_createLink',
        'FS_createDevice',
        'FS_unlink',
        'addRunDependency',
        'removeRunDependency',
        # from emscripten/tools/ctor_evaller.py
        '_sbrk', '___cxa_atexit', '_atexit', 'dumpGlobals',
        # from emscripten/system/lib/pthreads.symbols
        '___assert_fail', '___errno_location', '___lock',
        '___pthread_self', '___pthread_tsd_main',
        '___pthread_tsd_run_dtors', '___pthread_tsd_size',
        '___shm_mapname', '___timedwait', '___unlock',
        '___vm_lock', '___vm_lock_impl', '___vm_unlock',
        '___vm_unlock_impl', '___wait', '__pthread_getcanceltype',
        '__pthread_isduecanceled', '__pthread_msecs_until',
        '____atomic_is_lock_free', '__llvm_atomic_load_add_i32_p0i32',
        '__llvm_memory_barrier', '_accept', '_access',
        '_atexit', '_bind', '_calloc', '_chdir', '_chmod',
        '_chown', '_chroot', '_clearenv', '_clearerr',
        '_clock_gettime', '_close', '_closedir', '_confstr',
        '_connect', '_creat', '_crypt', '_ctermid', '_dprintf',
        '_dup', '_dup2', '_emscripten_asm_const',
        '_emscripten_futex_wait', '_emscripten_futex_wake',
        '_emscripten_get_now', '_emscripten_is_main_runtime_thread',
        '_emscripten_main_thread_process_queued_calls',
        '_emscripten_pthread_attr_settransferredcanvases',
        '_emscripten_pthread_attr_gettransferredcanvases',
        '_emscripten_sync_run_in_main_thread',
        '_emscripten_sync_run_in_main_thread_0',
        '_emscripten_sync_run_in_main_thread_1',
        '_emscripten_sync_run_in_main_thread_2',
        '_emscripten_sync_run_in_main_thread_3',
        '_emscripten_sync_run_in_main_thread_4',
        '_emscripten_sync_run_in_main_thread_5',
        '_emscripten_sync_run_in_main_thread_6',
        '_emscripten_sync_run_in_main_thread_xprintf_varargs',
        '_encrypt', '_fchdir', '_fchmod', '_fchown', '_fclose',
        '_fcntl', '_fdopen', '_feof', '_ferror', '_fflush',
        '_fgetc', '_fgetpos', '_fgets', '_fileno', '_fopen',
        '_fpathconf', '_fprintf', '_fputc', '_fputs', '_fread',
        '_free', '_freopen', '_fscanf', '_fseek', '_fsetpos',
        '_fstat', '_fstatvfs', '_fsync', '_ftell', '_ftruncate',
        '_fwrite', '_getchar', '_getcwd', '_getenv',
        '_gethostname', '_getlogin', '_getlogin_r',
        '_getpeername', '_gets', '_getsockname', '_getsockopt',
        '_gettimeofday', '_ioctl', '_isatty', '_lchmod',
        '_lchown', '_link', '_listen', '_lockf', '_lseek',
        '_lstat', '_malloc', '_mkdir', '_mkdtemp', '_mkfifo',
        '_mknod', '_mkstemp', '_mktemp', '_mmap', '_munmap',
        '_nanosleep', '_open', '_opendir', '_pclose',
        '_perror', '_pipe', '_poll', '_popen', '_posix_fallocate',
        '_pread', '_printf', '_pthread_attr_destroy',
        '_pthread_attr_getdetachstate', '_pthread_attr_getguardsize',
        '_pthread_attr_getinheritsched', '_pthread_attr_getschedparam',
        '_pthread_attr_getschedpolicy', '_pthread_attr_getscope',
        '_pthread_attr_getstack', '_pthread_attr_getstacksize',
        '_pthread_attr_init', '_pthread_attr_setdetachstate',
        '_pthread_attr_setguardsize', '_pthread_attr_setinheritsched',
        '_pthread_attr_setschedparam', '_pthread_attr_setschedpolicy',
        '_pthread_attr_setscope', '_pthread_attr_setstack',
        '_pthread_attr_setstacksize', '_pthread_barrier_destroy',
        '_pthread_barrier_init', '_pthread_barrier_wait',
        '_pthread_barrierattr_destroy',
        '_pthread_barrierattr_getpshared',
        '_pthread_barrierattr_init',
        '_pthread_barrierattr_setpshared',
        '_pthread_cleanup_pop',
        '_pthread_cleanup_push',
        '_pthread_cond_broadcast', '_pthread_cond_destroy',
        '_pthread_cond_init', '_pthread_cond_signal',
        '_pthread_cond_timedwait', '_pthread_cond_wait',
        '_pthread_condattr_destroy', '_pthread_condattr_getclock',
        '_pthread_condattr_getpshared', '_pthread_condattr_init',
        '_pthread_condattr_setclock',
        '_pthread_condattr_setpshared', '_pthread_create',
        '_pthread_equal', '_pthread_getattr_np',
        '_pthread_getspecific', '_pthread_key_create',
        '_pthread_key_delete', '_pthread_mutex_consistent',
        '_pthread_mutex_destroy', '_pthread_mutex_getprioceiling',
        '_pthread_mutex_init', '_pthread_mutex_lock',
        '_pthread_mutex_setprioceiling', '_pthread_mutex_timedlock',
        '_pthread_mutex_trylock', '_pthread_mutex_unlock',
        '_pthread_mutexattr_destroy',
        '_pthread_mutexattr_getprotocol',
        '_pthread_mutexattr_getpshared',
        '_pthread_mutexattr_getrobust',
        '_pthread_mutexattr_gettype', '_pthread_mutexattr_init',
        '_pthread_mutexattr_setprotocol',
        '_pthread_mutexattr_setpshared',
        '_pthread_mutexattr_setrobust',
        '_pthread_mutexattr_settype', '_pthread_once',
        '_pthread_rwlock_destroy', '_pthread_rwlock_init',
        '_pthread_rwlock_rdlock', '_pthread_rwlock_timedrdlock',
        '_pthread_rwlock_timedwrlock',
        '_pthread_rwlock_tryrdlock', '_pthread_rwlock_trywrlock',
        '_pthread_rwlock_unlock', '_pthread_rwlock_wrlock',
        '_pthread_rwlockattr_destroy',
        '_pthread_rwlockattr_getpshared',
        '_pthread_rwlockattr_init',
        '_pthread_rwlockattr_setpshared', '_pthread_self',
        '_pthread_setcancelstate', '_pthread_setcanceltype',
        '_pthread_setspecific', '_pthread_spin_destroy',
        '_pthread_spin_init', '_pthread_spin_lock',
        '_pthread_spin_trylock', '_pthread_spin_unlock',
        '_pthread_testcancel', '_putchar', '_putenv',
        '_puts', '_pwrite', '_read', '_readdir', '_readdir_r',
        '_readlink', '_realpath', '_recv', '_recvfrom',
        '_recvmsg', '_remove', '_rename', '_rewind', '_rewinddir',
        '_rmdir', '_sbrk', '_scanf', '_sched_get_priority_max',
        '_sched_get_priority_min', '_seekdir', '_select',
        '_sem_close', '_sem_destroy', '_sem_getvalue',
        '_sem_init', '_sem_open', '_sem_post', '_sem_timedwait',
        '_sem_trywait', '_sem_unlink', '_sem_wait', '_send',
        '_sendmsg', '_sendto', '_setenv', '_shm_unlink',
        '_shutdown', '_snprintf', '_socket', '_stat', '_statvfs',
        '_symlink', '_sysconf', '_tcgetattr', '_tcsetattr',
        '_telldir', '_tempnam', '_tmpfile', '_tmpnam',
        '_truncate', '_ttyname', '_ttyname_r', '_tzset',
        '_umask', '_ungetc', '_unlink', '_unsetenv', '_usleep',
        '_utime', '_utimes', '_vsnprintf', '_write']

    if x.startswith('_emscripten_'):
        return True
    # function_tables(...)
    elif x.startswith('dynCall_'):
        return True
    # create_basic_funcs(...)
    elif x.startswith('nullFunc_'):
        return True
    elif x.startswith('invoke_'):
        return True
    elif x.startswith('jsCall_'):
        return True
    elif x.startswith('ftCall_'):
        return True
    # syscalls
    elif x.replace('_', '').startswith('syscall'):
        return True
    elif x in EMSCRIPTEN_LIST:
        return True
    else:
        return False



# based on Manticore project: https://github.com/trailofbits/manticore
# extract from http://gavwood.com/paper.pdf

_table = {
    # opcode:(mnemonic, immediate_operand_size, pops, pushes, gas, description)
    0x00: ('STOP', 0, 0, 0, 0, 'Halts execution.'),
    0x01: ('ADD', 0, 2, 1, 3, 'Addition operation.'),
    0x02: ('MUL', 0, 2, 1, 5, 'Multiplication operation.'),
    0x03: ('SUB', 0, 2, 1, 3, 'Subtraction operation.'),
    0x04: ('DIV', 0, 2, 1, 5, 'Integer division operation.'),
    0x05: ('SDIV', 0, 2, 1, 5, 'Signed integer division operation (truncated).'),
    0x06: ('MOD', 0, 2, 1, 5, 'Modulo remainder operation.'),
    0x07: ('SMOD', 0, 2, 1, 5, 'Signed modulo remainder operation.'),
    0x08: ('ADDMOD', 0, 3, 1, 8, 'Modulo addition operation.'),
    0x09: ('MULMOD', 0, 3, 1, 8, 'Modulo multiplication operation.'),
    0x0a: ('EXP', 0, 2, 1, 10, 'Exponential operation.'),
    0x0b: ('SIGNEXTEND', 0, 2, 1, 5, "Extend length of two's complement signed integer."),
    0x10: ('LT', 0, 2, 1, 3, 'Less-than comparision.'),
    0x11: ('GT', 0, 2, 1, 3, 'Greater-than comparision.'),
    0x12: ('SLT', 0, 2, 1, 3, 'Signed less-than comparision.'),
    0x13: ('SGT', 0, 2, 1, 3, 'Signed greater-than comparision.'),
    0x14: ('EQ', 0, 2, 1, 3, 'Equality comparision.'),
    0x15: ('ISZERO', 0, 1, 1, 3, 'Simple not operator.'),
    0x16: ('AND', 0, 2, 1, 3, 'Bitwise AND operation.'),
    0x17: ('OR', 0, 2, 1, 3, 'Bitwise OR operation.'),
    0x18: ('XOR', 0, 2, 1, 3, 'Bitwise XOR operation.'),
    0x19: ('NOT', 0, 1, 1, 3, 'Bitwise NOT operation.'),
    0x1a: ('BYTE', 0, 2, 1, 3, 'Retrieve single byte from word.'),
    0x20: ('SHA3', 0, 2, 1, 30, 'Compute Keccak-256 hash.'),  # SHA3
    0x30: ('ADDRESS', 0, 0, 1, 2, 'Get address of currently executing account.'),
    0x31: ('BALANCE', 0, 1, 1, 20, 'Get balance of the given account.'),
    0x32: ('ORIGIN', 0, 0, 1, 2, 'Get execution origination address.'),
    0x33: ('CALLER', 0, 0, 1, 2, 'Get caller address.'),
    0x34: ('CALLVALUE', 0, 0, 1, 2, 'Get deposited value by the instruction/transaction responsible for this execution.'),
    0x35: ('CALLDATALOAD', 0, 1, 1, 3, 'Get input data of current environment.'),
    0x36: ('CALLDATASIZE', 0, 0, 1, 2, 'Get size of input data in current environment.'),
    0x37: ('CALLDATACOPY', 0, 3, 0, 3, 'Copy input data in current environment to memory.'),
    0x38: ('CODESIZE', 0, 0, 1, 2, 'Get size of code running in current environment.'),
    0x39: ('CODECOPY', 0, 3, 0, 3, 'Copy code running in current environment to memory.'),
    0x3a: ('GASPRICE', 0, 0, 1, 2, 'Get price of gas in current environment.'),
    0x3b: ('EXTCODESIZE', 0, 1, 1, 20, "Get size of an account's code."),
    0x3c: ('EXTCODECOPY', 0, 4, 0, 20, "Copy an account's code to memory."),
    0x3d: ('RETURNDATASIZE', 0, 0, 1, 2, "get size of return data buffer"),
    0x3e: ('RETURNDATACOPY', 0, 3, 0, 3, "copy return data in current environment to memory"),
    0x40: ('BLOCKHASH', 0, 1, 1, 20, 'Get the hash of one of the 256 most recent complete blocks.'),
    0x41: ('COINBASE', 0, 0, 1, 2, "Get the block's beneficiary address."),
    0x42: ('TIMESTAMP', 0, 0, 1, 2, "Get the block's timestamp."),
    0x43: ('NUMBER', 0, 0, 1, 2, "Get the block's number."),
    0x44: ('DIFFICULTY', 0, 0, 1, 2, "Get the block's difficulty."),
    0x45: ('GASLIMIT', 0, 0, 1, 2, "Get the block's gas limit."),
    0x50: ('POP', 0, 1, 0, 2, 'Remove item from stack.'),
    0x51: ('MLOAD', 0, 1, 1, 3, 'Load word from memory.'),
    0x52: ('MSTORE', 0, 2, 0, 3, 'Save word to memory.'),
    0x53: ('MSTORE8', 0, 2, 0, 3, 'Save byte to memory.'),
    0x54: ('SLOAD', 0, 1, 1, 50, 'Load word from storage.'),
    0x55: ('SSTORE', 0, 2, 0, 0, 'Save word to storage.'),
    0x56: ('JUMP', 0, 1, 0, 8, 'Alter the program counter.'),
    0x57: ('JUMPI', 0, 2, 0, 10, 'Conditionally alter the program counter.'),
    0x58: ('GETPC', 0, 0, 1, 2, 'Get the value of the program counter prior to the increment.'),
    0x59: ('MSIZE', 0, 0, 1, 2, 'Get the size of active memory in bytes.'),
    0x5a: ('GAS', 0, 0, 1, 2, 'Get the amount of available gas, including the corresponding reduction the amount of available gas.'),
    0x5b: ('JUMPDEST', 0, 0, 0, 1, 'Mark a valid destination for jumps.'),
    0x60: ('PUSH1', 1, 0, 1, 0, 'Place 1 byte item on stack.'),
    0x61: ('PUSH2', 2, 0, 1, 0, 'Place 2-byte item on stack.'),
    0x62: ('PUSH3', 3, 0, 1, 0, 'Place 3-byte item on stack.'),
    0x63: ('PUSH4', 4, 0, 1, 0, 'Place 4-byte item on stack.'),
    0x64: ('PUSH5', 5, 0, 1, 0, 'Place 5-byte item on stack.'),
    0x65: ('PUSH6', 6, 0, 1, 0, 'Place 6-byte item on stack.'),
    0x66: ('PUSH7', 7, 0, 1, 0, 'Place 7-byte item on stack.'),
    0x67: ('PUSH8', 8, 0, 1, 0, 'Place 8-byte item on stack.'),
    0x68: ('PUSH9', 9, 0, 1, 0, 'Place 9-byte item on stack.'),
    0x69: ('PUSH10', 10, 0, 1, 0, 'Place 10-byte item on stack.'),
    0x6a: ('PUSH11', 11, 0, 1, 0, 'Place 11-byte item on stack.'),
    0x6b: ('PUSH12', 12, 0, 1, 0, 'Place 12-byte item on stack.'),
    0x6c: ('PUSH13', 13, 0, 1, 0, 'Place 13-byte item on stack.'),
    0x6d: ('PUSH14', 14, 0, 1, 0, 'Place 14-byte item on stack.'),
    0x6e: ('PUSH15', 15, 0, 1, 0, 'Place 15-byte item on stack.'),
    0x6f: ('PUSH16', 16, 0, 1, 0, 'Place 16-byte item on stack.'),
    0x70: ('PUSH17', 17, 0, 1, 0, 'Place 17-byte item on stack.'),
    0x71: ('PUSH18', 18, 0, 1, 0, 'Place 18-byte item on stack.'),
    0x72: ('PUSH19', 19, 0, 1, 0, 'Place 19-byte item on stack.'),
    0x73: ('PUSH20', 20, 0, 1, 0, 'Place 20-byte item on stack.'),
    0x74: ('PUSH21', 21, 0, 1, 0, 'Place 21-byte item on stack.'),
    0x75: ('PUSH22', 22, 0, 1, 0, 'Place 22-byte item on stack.'),
    0x76: ('PUSH23', 23, 0, 1, 0, 'Place 23-byte item on stack.'),
    0x77: ('PUSH24', 24, 0, 1, 0, 'Place 24-byte item on stack.'),
    0x78: ('PUSH25', 25, 0, 1, 0, 'Place 25-byte item on stack.'),
    0x79: ('PUSH26', 26, 0, 1, 0, 'Place 26-byte item on stack.'),
    0x7a: ('PUSH27', 27, 0, 1, 0, 'Place 27-byte item on stack.'),
    0x7b: ('PUSH28', 28, 0, 1, 0, 'Place 28-byte item on stack.'),
    0x7c: ('PUSH29', 29, 0, 1, 0, 'Place 29-byte item on stack.'),
    0x7d: ('PUSH30', 30, 0, 1, 0, 'Place 30-byte item on stack.'),
    0x7e: ('PUSH31', 31, 0, 1, 0, 'Place 31-byte item on stack.'),
    0x7f: ('PUSH32', 32, 0, 1, 0, 'Place 32-byte (full word) item on stack.'),
    0x80: ('DUP1', 0, 1, 2, 3, 'Duplicate 1st stack item.'),
    0x81: ('DUP2', 0, 2, 3, 3, 'Duplicate 2nd stack item.'),
    0x82: ('DUP3', 0, 3, 4, 3, 'Duplicate 3rd stack item.'),
    0x83: ('DUP4', 0, 4, 5, 3, 'Duplicate 4th stack item.'),
    0x84: ('DUP5', 0, 5, 6, 3, 'Duplicate 5th stack item.'),
    0x85: ('DUP6', 0, 6, 7, 3, 'Duplicate 6th stack item.'),
    0x86: ('DUP7', 0, 7, 8, 3, 'Duplicate 7th stack item.'),
    0x87: ('DUP8', 0, 8, 9, 3, 'Duplicate 8th stack item.'),
    0x88: ('DUP9', 0, 9, 10, 3, 'Duplicate 9th stack item.'),
    0x89: ('DUP10', 0, 10, 11, 3, 'Duplicate 10th stack item.'),
    0x8a: ('DUP11', 0, 11, 12, 3, 'Duplicate 11th stack item.'),
    0x8b: ('DUP12', 0, 12, 13, 3, 'Duplicate 12th stack item.'),
    0x8c: ('DUP13', 0, 13, 14, 3, 'Duplicate 13th stack item.'),
    0x8d: ('DUP14', 0, 14, 15, 3, 'Duplicate 14th stack item.'),
    0x8e: ('DUP15', 0, 15, 16, 3, 'Duplicate 15th stack item.'),
    0x8f: ('DUP16', 0, 16, 17, 3, 'Duplicate 16th stack item.'),
    0x90: ('SWAP1', 0, 2, 2, 3, 'Exchange 1st and 2nd stack items.'),
    0x91: ('SWAP2', 0, 3, 3, 3, 'Exchange 1st and 3rd stack items.'),
    0x92: ('SWAP3', 0, 4, 4, 3, 'Exchange 1st and 4th stack items.'),
    0x93: ('SWAP4', 0, 5, 5, 3, 'Exchange 1st and 5th stack items.'),
    0x94: ('SWAP5', 0, 6, 6, 3, 'Exchange 1st and 6th stack items.'),
    0x95: ('SWAP6', 0, 7, 7, 3, 'Exchange 1st and 7th stack items.'),
    0x96: ('SWAP7', 0, 8, 8, 3, 'Exchange 1st and 8th stack items.'),
    0x97: ('SWAP8', 0, 9, 9, 3, 'Exchange 1st and 9th stack items.'),
    0x98: ('SWAP9', 0, 10, 10, 3, 'Exchange 1st and 10th stack items.'),
    0x99: ('SWAP10', 0, 11, 11, 3, 'Exchange 1st and 11th stack items.'),
    0x9a: ('SWAP11', 0, 12, 12, 3, 'Exchange 1st and 12th stack items.'),
    0x9b: ('SWAP12', 0, 13, 13, 3, 'Exchange 1st and 13th stack items.'),
    0x9c: ('SWAP13', 0, 14, 14, 3, 'Exchange 1st and 14th stack items.'),
    0x9d: ('SWAP14', 0, 15, 15, 3, 'Exchange 1st and 15th stack items.'),
    0x9e: ('SWAP15', 0, 16, 16, 3, 'Exchange 1st and 16th stack items.'),
    0x9f: ('SWAP16', 0, 17, 17, 3, 'Exchange 1st and 17th stack items.'),
    0xa0: ('LOG0', 0, 2, 0, 375, 'Append log record with no topics.'),
    0xa1: ('LOG1', 0, 3, 0, 750, 'Append log record with one topic.'),
    0xa2: ('LOG2', 0, 4, 0, 1125, 'Append log record with two topics.'),
    0xa3: ('LOG3', 0, 5, 0, 1500, 'Append log record with three topics.'),
    0xa4: ('LOG4', 0, 6, 0, 1875, 'Append log record with four topics.'),
    0xf0: ('CREATE', 0, 3, 1, 32000, 'Create a new account with associated code.'),
    0xf1: ('CALL', 0, 7, 1, 40, 'Message-call into an account.'),
    0xf2: ('CALLCODE', 0, 7, 1, 40, "Message-call into this account with alternative account's code."),
    0xf3: ('RETURN', 0, 2, 0, 0, 'Halt execution returning output data.'),
    0xf4: ('DELEGATECALL', 0, 6, 1, 40, "like CALLCODE but keeps caller's value and sender"),
    0xfa: ('STATICCALL', 0, 6, 1, 40, 'like CALL but disallow state modifications'),
    0xfb: ('CREATE2', 0, 4, 1, 32000, 'create new account with associated code at address `sha3(sender + salt + sha3(init code)) % 2**160`'),
    0xfd: ('REVERT', 0, 2, 0, 0, 'halt execution, revert state and return output data'),
    0xfe: ('INVALID', 0, 0, 0, 0, 'old ASSERTFAIL - invalid instruction for expressing runtime errors (e.g., division-by-zero)'),
    0xff: ('SELFDESTRUCT', 0, 1, 0, 5000, 'Halt execution and register account for later deletion.')
}


class EVM(object):
    """Bytecode for Ethereum VM."""

    def __init__(self):
        self.table = _table
        self.reverse_table = self._get_reverse_table()

    def _get_reverse_table(self):
        """Build an internal table used in the assembler."""
        reverse_table = {}
        for (opcode, (mnemonic, immediate_operand_size,
                      pops, pushes, gas, description)) in _table.items():
            reverse_table[mnemonic] = opcode, mnemonic, immediate_operand_size, \
                pops, pushes, gas, description
        return reverse_table



from octopus.engine.engine import VMstate


class WasmVMstate(VMstate):

    def __init__(self):

        self.memory = []

        self.stack = []
        self.ssa_stack = []
        self.symbolic_stack = []

        self.last_returned = []
        self.pc = 0
        self.instr = None

        self.instructions_visited = list()
        #self.instructions_visited = dict()

    def details(self):

        return {'memory': self.memory,
                'stack': self.stack,
                'ssa_stack': self.ssa_stack,
                'symbolic_stack': self.symbolic_stack,
                'last_returned': self.last_returned,
                'pc': self.pc}

    def mem_extend(self, start, sz):

        if (start < 4096 and sz < 4096):

            if sz and start + sz > len(self.memory):

                n_append = start + sz - len(self.memory)

                while n_append > 0:
                    self.memory.append(0)
                    n_append -= 1

        else:
            raise Exception('mem_extend')



from octopus.engine.disassembler import Disassembler
from octopus.core.function import Function
from octopus.core.utils import bytecode_to_bytes

from octopus.arch.wasm.instruction import WasmInstruction
from octopus.arch.wasm.wasm import Wasm

from octopus.arch.wasm.decode import decode_module
# from wasm.decode import decode_module
from wasm.modtypes import CodeSection
from wasm.compat import byte2int
from wasm.opcodes import OPCODE_MAP
from wasm.formatter import format_instruction

from collections import namedtuple
inst_namedtuple = namedtuple('Instruction', 'op imm len')


class WasmDisassembler(Disassembler):

    def __init__(self, bytecode=None):
        Disassembler.__init__(self, asm=Wasm(), bytecode=bytecode)

    def disassemble_opcode(self, bytecode=None, offset=0):
        '''
        based on decode_bytecode()
        https://github.com/athre0z/wasm/blob/master/wasm/decode.py

        '''

        bytecode_wnd = memoryview(bytecode)
        opcode_id = byte2int(bytecode_wnd[0])

        # default value
        # opcode:(mnemonic/name, imm_struct, pops, pushes, description)
        invalid = ('INVALID', 0, 0, 0, 'Unknown opcode')
        name, imm_struct, pops, pushes, description = \
            self.asm.table.get(opcode_id, invalid)

        operand_size = 0
        operand = None
        operand_interpretation = None

        if imm_struct is not None:
            operand_size, operand, _ = imm_struct.from_raw(None, bytecode_wnd[1:])
            insn = inst_namedtuple(OPCODE_MAP[opcode_id], operand, 1 + operand_size)
            operand_interpretation = format_instruction(insn)
        insn_byte = bytecode_wnd[:1 + operand_size].tobytes()
        instruction = WasmInstruction(opcode_id, name, imm_struct, operand_size,
                                      insn_byte, pops, pushes, description,
                                      operand_interpretation=operand_interpretation,
                                      offset=offset)
        # print('%d %s' % (offset, str(instruction)))
        return instruction

    def disassemble(self, bytecode=None, offset=0, r_format='list'):
        """Disassemble WASM bytecode

        :param bytecode: bytecode sequence
        :param offset: start offset
        :param r_format: output format ('list'/'text'/'reverse')
        :type bytecode: bytes, str
        :type offset: int
        :type r_format: list, str, dict
        :return: dissassembly result depending of r_format
        :rtype: list, str, dict

        :Example:

        >>> disasm = WasmDisassembler()
        >>>
        >>> disasm.disassemble(r_format='text')
        >>> 'block -1\ni32.const 24\ncall 28\ni32.const 0\nreturn\nend'
        >>>
        >>> disasm.disassemble(r_format='text')
        >>> [<octopus.arch.wasm.instruction.WasmInstruction at 0x7f80243120b8>,
             ...
             <octopus.arch.wasm.instruction.WasmInstruction at 0x7f8024312588>,
             <octopus.arch.wasm.instruction.WasmInstruction at 0x7f80243121d0>]
        >>>
        >>> disasm.disassemble(r_format='reverse')
        >>> {0: <octopus.arch.wasm.instruction.WasmInstruction at 0x7f8024319d68>,
             ...
             4: <octopus.arch.wasm.instruction.WasmInstruction at 0x7f802431fa58>,
             5: <octopus.arch.wasm.instruction.WasmInstruction at 0x7f802431fc18>}
        """

        return super().disassemble(bytecode, offset, r_format)

    def extract_functions_code(self, module_bytecode):
        functions = list()
        mod_iter = iter(decode_module(module_bytecode))
        _, _ = next(mod_iter)
        sections = list(mod_iter)

        # iterate over all section
        #code_data = [cur_sec_data for cur_sec, cur_sec_data in sections if isinstance(cur_sec_data.get_decoder_meta()['types']['payload'], CodeSection)][0]
        for cur_sec, cur_sec_data in sections:
            sec = cur_sec_data.get_decoder_meta()['types']['payload']
            if isinstance(sec, CodeSection):
                code_data = cur_sec_data
                break
        if not code_data:
            raise ValueError('No functions/codes in the module')
        for idx, func in enumerate(code_data.payload.bodies):
            instructions = self.disassemble(func.code.tobytes())
            cur_function = Function(0, instructions[0])
            cur_function.instructions = instructions

            functions.append(cur_function)
        return functions

    def disassemble_module(self, module_bytecode=None, offset=0, r_format='list'):

        bytecode = bytecode_to_bytes(module_bytecode)

        functions = self.extract_functions_code(bytecode[offset:])
        self.instructions = [f.instructions for f in functions]

        # return instructions
        if r_format == 'list':
            return self.instructions
        elif r_format == 'text':
            text = ''
            for index, func in enumerate(functions):
                text += ('func %d\n' % index)
                text += ('\n'.join(map(str, func.instructions)))
                text += ('\n\n')
            return text



import unittest

from octopus.platforms.EOS.cfg import EosCFG
from octopus.analysis.graph import CallGraph


class EosCallGraphTestCase(unittest.TestCase):

    def testCallGraph(self):
        def callgraph(bytecode, result_nodes, length_edges, fname=False):
            cfg = EosCFG(bytecode)
            nodes, edges = cfg.get_functions_call_edges(fname)
            # visualize
            # graph = CallGraph(nodes, edges)
            # graph.view()
            self.assertEqual(nodes, result_nodes)
            self.assertEqual(len(edges), length_edges)

        # Helloworld
        bytecode_hex = "0061736d0100000001110460017f0060017e0060000060027e7e00021b0203656e76067072696e746e000103656e76067072696e7473000003030202030404017000000503010001071903066d656d6f7279020004696e69740002056170706c7900030a20020600411010010b17004120100120001000413010012001100041c00010010b0b3f050041040b04504000000041100b0d496e697420576f726c64210a000041200b0e48656c6c6f20576f726c643a20000041300b032d3e000041c0000b020a000029046e616d6504067072696e746e0100067072696e7473010004696e697400056170706c790201300131"
        r_nodes = ['printn(i64)', 'prints(i32)', 'init()', 'apply(i64 i64)']
        r_nodes2 = ['printn', 'prints', 'init', 'apply']
        len_edges = 6
        callgraph(bytecode_hex, r_nodes2, len_edges)
        callgraph(bytecode_hex, r_nodes, len_edges, fname=True)

if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(EosCallGraphTestCase)
    unittest.TextTestRunner(verbosity=2).run(suite)



from octopus.core.function import Function
from octopus.core.basicblock import BasicBlock
from octopus.core.edge import Edge
from octopus.core.edge import (EDGE_UNCONDITIONAL,
                               EDGE_CONDITIONAL_TRUE,
                               EDGE_CONDITIONAL_FALSE,
                               EDGE_FALLTHROUGH)

from octopus.analysis.cfg import CFG

from octopus.platforms.NEO.disassembler import NeoDisassembler

from logging import getLogger
logging = getLogger(__name__)


def enum_func_static(instructions):

    functions = list()

    # first function of NeoContract is *usually* the Main function
    function = Function(instructions[0].offset, start_instr=instructions[0], name='Main')

    # parse the instructions and create Function object
    new_func = False
    for inst in instructions:
        if new_func:
            name = 'func_%x' % inst.offset
            function = Function(inst.offset, start_instr=inst, name=name)
            new_func = False

        # associate current instruction to the function
        # Worked because instruction between
        # function.start_offset and function.end_offset
        # are on the same function on Neo
        # i.e linear disassembly
        function.instructions.append(inst)

        # terminator instruction or last one
        if inst.is_halt or inst == instructions[-1]:
            function.size = inst.offset_end - function.start_offset
            function.end_offset = inst.offset_end
            function.end_instr = inst

            functions.append(function)
            new_func = True

    return functions


def xref_of_instr(instr):
    """ Return xref for branch instruction """
    return int.from_bytes(instr.operand, byteorder='little', signed=True) + instr.offset_end - 2


def enumerate_xref(instructions):
    # list cross reference to Instruction and return list of xref
    xrefs = []

    for inst in instructions:
        if inst.is_branch_unconditional:
            xrefs.append(xref_of_instr(inst))
        elif inst.is_branch_conditional:
            xrefs.append(xref_of_instr(inst))
            xrefs.append(inst.offset_end + 1)
        elif inst.is_halt and inst != instructions[-1]:
            xrefs.append(inst.offset_end + 1)
    xrefs = list(set(xrefs))
    return xrefs


def assign_blocks_to_func(basicblocks, functions):

    for func in functions:
        for bb in basicblocks:
            if bb.start_offset in range(func.start_offset, func.end_offset):
                func.basicblocks.append(bb)
    return functions


def enum_blocks_edges(instructions):

    """
    Return a list of basicblock after
    statically parsing given instructions
    """

    basicblocks = list()
    edges = list()
    xrefs = enumerate_xref(instructions)
    # create the first block
    new_block = True

    for inst in instructions:
        if new_block:
            block = BasicBlock(start_offset=inst.offset,
                               start_instr=inst,
                               name='block_%x' % inst.offset)
            new_block = False

        # add current instruction to the basicblock
        block.instructions.append(inst)

        # next instruction in xrefs list
        if (inst.offset_end + 1) in xrefs:
            # absolute JUMP
            if inst.is_branch_unconditional:
                edges.append(Edge(block.name, 'block_%x' % xref_of_instr(inst),
                                  EDGE_UNCONDITIONAL))
            # conditionnal JUMPI / JUMPIF / ...
            elif inst.is_branch_conditional:
                edges.append(Edge(block.name, 'block_%x' % xref_of_instr(inst),
                                  EDGE_CONDITIONAL_TRUE))
                edges.append(Edge(block.name, 'block_%x' % (inst.offset_end + 1),
                                  EDGE_CONDITIONAL_FALSE))
            # Halt instruction : RETURN, STOP, RET, ...
            elif inst.is_halt:
                pass
            # just falls to the next instruction
            else:
                edges.append(Edge(block.name, 'block_%x' % (inst.offset_end + 1),
                                  EDGE_FALLTHROUGH))

            block.end_offset = inst.offset_end
            block.end_instr = inst
            basicblocks.append(block)
            new_block = True

    # add the last block
    basicblocks.append(block)
    edges = list(set(edges))

    return (basicblocks, edges)


class NeoCFG(CFG):

    def __init__(self, bytecode=None, instructions=None, static_analysis=True):
        """ TODO """

        if not bytecode and not instructions:
            raise Exception("No bytecode/instructions provided")
        if instructions:
            self.instructions = instructions
        else:
            # disassemble bytecode to instructions
            disasm = NeoDisassembler(bytecode)
            self.instructions = disasm.disassemble()
        self.static_analysis = static_analysis

        self.basicblocks = list()
        self.functions = list()
        self.edges = list()

        if self.static_analysis:
            self.run_static_analysis()

    def run_static_analysis(self):
        self.functions = enum_func_static(self.instructions)
        self.basicblocks, self.edges = enum_blocks_edges(self.instructions)
        self.functions = assign_blocks_to_func(self.basicblocks, self.functions)

    def __str__(self):
        line = ("length functions = %d\n" % len(self.functions))
        line += ("length basicblocks = %d\n" % len(self.basicblocks))
        line += ("length instructions = %d\n" % len(self.instructions))
        line += ("length edges = %d\n" % len(self.edges))
        return line



# =======================================
# #     Symbolic Execution Engine       #
# =======================================


class SymbolicExecutionEngine(object):

    def __init__(self, bytecode=None, instructions=None, contract=None, modules=None, dynamic_loader=None, max_depth=10):
        """ TODO """
        raise NotImplementedError

    def sym_exec_contract(self, contract):
        """ TODO """
        raise NotImplementedError

    def sym_exec_single_instr(self, instr, context, state, depth, constraints):
        """ TODO """
        raise NotImplementedError
